unit Dialogs;

{$I define.pp}

interface

uses
{ Trace, }
 Objects, Drivers, Views;

const

{ Color palettes }

  CGrayDialog    = #32#33#34#35#36#37#38#39#40#41#42#43#44#45#46#47 +
                   #48#49#50#51#52#53#54#55#56#57#58#59#60#61#62#63;
  CBlueDialog    = #64#65#66#67#68#69#70#71#72#73#74#75#76#77#78#79 +
                   #80#81#82#83#84#85#86#87#88#89#90#91#92#92#94#95;
  CCyanDialog    = #96#97#98#99#100#101#102#103#104#105#106#107#108 +
                   #109#110#111#112#113#114#115#116#117#118#119#120 +
                   #121#122#123#124#125#126#127;

  CDialog        = CGrayDialog;

  CStaticText    = #6;
  CLabel         = #7#8#9#9;
  CButton        = #10#11#12#13#14#14#14#15;
  CCluster       = #16#17#18#18#31;
  CInputLine     = #19#19#20#21;
  CHistory       = #22#23;
  CHistoryWindow = #19#19#21#24#25#19#20;
  CHistoryViewer = #6#6#7#6#6;

{ TDialog palette entires }

  dpBlueDialog = 0;
  dpCyanDialog = 1;
  dpGrayDialog = 2;

{ TButton flags }

  bfNormal    = $00;
  bfDefault   = $01;
  bfLeftJust  = $02;
  bfBroadcast = $04;
  bfGrabFocus = $08;

type

{ TDialog object }

  { Palette layout }
  {  1 = Frame passive }
  {  2 = Frame active }
  {  3 = Frame icon }
  {  4 = ScrollBar page area }
  {  5 = ScrollBar controls }
  {  6 = StaticText }
  {  7 = Label normal }
  {  8 = Label selected }
  {  9 = Label shortcut }
  { 10 = Button normal }
  { 11 = Button default }
  { 12 = Button selected }
  { 13 = Button disabled }
  { 14 = Button shortcut }
  { 15 = Button shadow }
  { 16 = Cluster normal }
  { 17 = Cluster selected }
  { 18 = Cluster shortcut }
  { 19 = InputLine normal text }
  { 20 = InputLine selected text }
  { 21 = InputLine arrows }
  { 22 = History arrow }
  { 23 = History sides }
  { 24 = HistoryWindow scrollbar page area }
  { 25 = HistoryWindow scrollbar controls }
  { 26 = ListViewer normal }
  { 27 = ListViewer focused }
  { 28 = ListViewer selected }
  { 29 = ListViewer divider }
  { 30 = InfoPane }
  { 31 = Cluster disabled }
  { 32 = Reserved }

  PDialog = ^TDialog;
  TDialog = object(TWindow)
    constructor Init(var Bounds: TRect; ATitle: TTitleStr);
    function GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    function Valid(Command: Word): Boolean; virtual;
  end;

{ TInputLine object }

  { Palette layout }
  { 1 = Passive }
  { 2 = Active }
  { 3 = Selected }
  { 4 = Arrows }

  PInputLine = ^TInputLine;
  TInputLine = object(TView)
    Data: PString;
    MaxLen: Integer;
    CurPos: Integer;
    FirstPos: Integer;
    SelStart: Integer;
    SelEnd: Integer;
    constructor Init(var Bounds: TRect; AMaxLen: Integer);
    destructor Done; virtual;
    function DataSize: Word; virtual;
    procedure Draw; virtual;
    procedure GetData(var Rec); virtual;
    function GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure SelectAll(Enable: Boolean);
    procedure SetData(var Rec); virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    function Valid(Command: Word): Boolean; virtual;
  private
    function CanScroll(Delta: Integer): Boolean;
  end;

{ TButton object }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Default text }
  { 3 = Selected text }
  { 4 = Disabled text }
  { 5 = Normal shortcut }
  { 6 = Default shortcut }
  { 7 = Selected shortcut }
  { 8 = Shadow }

  PButton = ^TButton;
  TButton = object(TView)
    Title: PString;
    Command: Word;
    Flags: Byte;
    AmDefault: Boolean;
    constructor Init(var Bounds: TRect; ATitle: TTitleStr; ACommand: Word;
      AFlags: Word);
    destructor Done; virtual;
    procedure Draw; virtual;
    procedure DrawState(Down: Boolean);
    function GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure MakeDefault(Enable: Boolean);
    procedure Press; virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
  end;

{ TCheckBox }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }
  { 3 = Normal shortcut }
  { 4 = Selected shortcut }
  { 5 = Disabled text }

  PCheckBox = ^TCheckBox;
  TCheckBox = object(TView)
    Value:boolean;
    constructor Init(var Bounds: TRect);
    destructor Done; virtual;
    function DataSize: Word; virtual;
    procedure Draw; virtual;
{
    procedure DrawMultiBox(const Icon, Marker: String);
}
    procedure GetData(var Rec); virtual;
    function GetHelpCtx: Word; virtual;
    function GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    function Mark: Boolean; virtual;
{
    function MultiMark: Byte; virtual;
}
    procedure Press; virtual;
    procedure SetData(var Rec); virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
  private
    function FindSel(P: TPoint): boolean;
  end;

{ TStaticText }

  { Palette layout }
  { 1 = Text }

  PStaticText = ^TStaticText;
  TStaticText = object(TView)
    Text: PString;
    constructor Init(var Bounds: TRect; const AText: String);
    destructor Done; virtual;
    procedure Draw; virtual;
    function GetPalette: PPalette; virtual;
    procedure GetText(var S: String); virtual;
  end;

{ TParamText }

  { Palette layout }
  { 1 = Text }

  PParamText = ^TParamText;
  TParamText = object(TStaticText)
    ParamCount: Integer;
    ParamList: Pointer;
    constructor Init(var Bounds: TRect; const AText: String;
      AParamCount: Integer);
    function DataSize: Word; virtual;
    procedure GetText(var S: String); virtual;
    procedure SetData(var Rec); virtual;
  end;

{ TLabel }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }
  { 3 = Normal shortcut }
  { 4 = Selected shortcut }

  PLabel = ^TLabel;
  TLabel = object(TStaticText)
    Link: PView;
    Light: Boolean;
    constructor Init(var Bounds: TRect; const AText: String; ALink: PView);
    procedure Draw; virtual;
    function GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
  end;

const

{ Dialog broadcast commands }

  cmRecordHistory = 60;

implementation

const

{ TButton messages }

  cmGrabDefault    = 61;
  cmReleaseDefault = 62;

{ Utility functions }

function IsBlank(Ch: Char): Boolean;
 begin IsBlank := (Ch = ' ') or (Ch = #13) or (Ch = #10) end;

{ TDialog }

constructor TDialog.Init(var Bounds: TRect; ATitle: TTitleStr);
 begin
  inherited Init(Bounds, ATitle, wnNoNumber);
  GrowMode := 0;
  Flags := wfMove + wfClose;
  Palette := dpGrayDialog;
 end;

function TDialog.GetPalette: PPalette;
 const
  P: array[dpBlueDialog..dpGrayDialog] of string[Length(CBlueDialog)] =
    (CBlueDialog, CCyanDialog, CGrayDialog);
 begin GetPalette := @P[Palette] end;

procedure TDialog.HandleEvent(var Event: TEvent);
 begin
  TWindow.HandleEvent(Event);
  case Event.What of
    evKeyDown:
      case Event.KeyCode of
        kbEsc:
          begin
            Event.What := evCommand;
            Event.Command := cmCancel;
            Event.InfoPtr := nil;
            PutEvent(Event);
            ClearEvent(Event);
          end;
        kbEnter:
          begin
            Event.What := evBroadcast;
            Event.Command := cmDefault;
            Event.InfoPtr := nil;
            PutEvent(Event);
            ClearEvent(Event);
          end;
      end;
    evCommand:
      case Event.Command of
        cmOk, cmCancel, cmYes, cmNo:
          if State and sfModal <> 0 then
          begin
            EndModal(Event.Command);
            ClearEvent(Event);
          end;
      end;
  end;
 end;

function TDialog.Valid(Command: Word): Boolean;
 begin
  if Command = cmCancel
   then Valid := True
   else Valid := TGroup.Valid(Command);
 end;

function Max(A, B: Integer): Integer;
 begin if a>b then max:=a else max:=b end;

function HotKey(const S: String): Char;
 var P: Word;
 begin
  P := Pos('~',S);
  if P <> 0 then HotKey := UpCase(S[P+1])
  else HotKey := #0;
 end;

{ TInputLine }

constructor TInputLine.Init(var Bounds: TRect; AMaxLen: Integer);
 begin
  TView.Init(Bounds);
  State := State or sfCursorVis;
  Options := Options or (ofSelectable + ofFirstClick);
  GetMem(Data, AMaxLen + 1);
  Data^ := ''; MaxLen := AMaxLen;
 end;

destructor TInputLine.Done;
 begin
  FreeMem(Data, MaxLen + 1);
  TView.Done;
 end;

function TInputLine.CanScroll(Delta: Integer): Boolean;
 begin
  if Delta < 0 then CanScroll := FirstPos > 0 else
   if Delta > 0 then CanScroll := Length(Data^) - FirstPos + 2 > Size.X else
    CanScroll := False;
 end;

function TInputLine.DataSize: Word;
 begin DataSize := MaxLen+1 end;

procedure TInputLine.Draw;
 var Color: Byte; L, R: Integer; B: TDrawBuffer;
 begin
  if State and sfFocused = 0 then
    Color := GetColor(1) else
    Color := GetColor(2);
  MoveChar(B, ' ', Color, Size.X);
  MoveStr(B[1], Copy(Data^, FirstPos + 1, Size.X - 2), Color);
  if CanScroll(1) then MoveChar(B[Size.X - 1], #16, GetColor(4), 1);
  if State and sfFocused <> 0 then
  begin
    if CanScroll(-1) then MoveChar(B[0], #17, GetColor(4), 1);
    L := SelStart - FirstPos;
    R := SelEnd - FirstPos;
    if L < 0 then L := 0;
    if R > Size.X - 2 then R := Size.X - 2;
    if L < R then MoveChar(B[L + 1], #0, GetColor(3), R - L);
  end;
  WriteLine(0, 0, Size.X, Size.Y, B);
  SetCursor(CurPos - FirstPos + 1, 0);
 end;

procedure TInputLine.GetData(var Rec);
 begin
  FillChar(Rec, DataSize, #0);
  Move(Data^, Rec, Length(Data^) + 1);
 end;

function TInputLine.GetPalette: PPalette;
 const P: String[Length(CInputLine)] = CInputLine;
 begin GetPalette := @P end;

procedure TInputLine.HandleEvent(var Event: TEvent);
 const PadKeys = [$47, $4B, $4D, $4F, $73, $74];
 var Delta, Anchor, I: Integer; ExtendBlock: Boolean;
 function MouseDelta: Integer;
  var Mouse: TPoint;
  begin
   MakeLocal(Event.Where, Mouse);
   if Mouse.X <= 0 then MouseDelta := -1 else
   if Mouse.X >= Size.X - 1 then MouseDelta := 1 else
   MouseDelta := 0;
  end;
 function MousePos: Integer;
  var Pos: Integer; Mouse: TPoint;
  begin
   MakeLocal(Event.Where, Mouse);
   if Mouse.X < 1 then Mouse.X := 1;
   Pos := Mouse.X + FirstPos - 1;
   if Pos < 0 then Pos := 0;
   if Pos > Length(Data^) then Pos := Length(Data^);
   MousePos := Pos;
  end;
 procedure DeleteSelect;
  begin if SelStart <> SelEnd then begin
    Delete(Data^, SelStart + 1, SelEnd - SelStart);
    CurPos := SelStart;
  end end;
 procedure AdjustSelectBlock;
  begin if CurPos < Anchor
   then begin SelStart := CurPos; SelEnd := Anchor end
   else begin SelStart := Anchor; SelEnd := CurPos end
  end;
 begin
  TView.HandleEvent(Event);
  if State and sfSelected <> 0 then begin
   case Event.What of
    evMouseDown:begin
     Delta := MouseDelta;
     if CanScroll(Delta) then begin
      repeat
       if CanScroll(Delta)
        then begin Inc(FirstPos, Delta); DrawView end;
      until not MouseEvent(Event, evMouseAuto);
     end else if Event.Double then SelectAll(True) else begin
      Anchor := MousePos;
      repeat
       if Event.What = evMouseAuto then begin
        Delta := MouseDelta;
        if CanScroll(Delta) then Inc(FirstPos, Delta);
       end;
       CurPos := MousePos;
       AdjustSelectBlock;
       DrawView;
      until not MouseEvent(Event, evMouseMove + evMouseAuto);
     end;
     ClearEvent(Event);
    end;
    evKeyDown:begin
     Event.KeyCode := CtrlToArrow(Event.KeyCode);
     if (Event.ScanCode in PadKeys) and
        (GetShiftState and $03 <> 0) then begin
      Event.CharCode := #0;
      if CurPos = SelEnd then Anchor := SelStart else Anchor := SelEnd;
      ExtendBlock := True;
     end else ExtendBlock := False;
     case Event.KeyCode of
      kbLeft:if CurPos > 0 then Dec(CurPos);
      kbRight:if CurPos < Length(Data^) then Inc(CurPos);
      kbHome:CurPos := 0;
      kbEnd:CurPos := Length(Data^);
      kbBack:if CurPos > 0 then begin
       Delete(Data^, CurPos, 1); Dec(CurPos);
       if FirstPos > 0 then Dec(FirstPos);
      end;
      kbDel:begin
       if SelStart = SelEnd then if CurPos < Length(Data^)
        then begin SelStart := CurPos; SelEnd := CurPos + 1 end;
       DeleteSelect;
      end;
      kbIns:SetState(sfCursorIns, State and sfCursorIns = 0);
      else case Event.CharCode of
       ' '..#255:begin
        if State and sfCursorIns <> 0
         then Delete(Data^, CurPos + 1, 1)
         else DeleteSelect;
        if Length(Data^) < MaxLen then begin
         if FirstPos > CurPos then FirstPos := CurPos;
         Inc(CurPos);
         Insert(Event.CharCode, Data^, CurPos);
        end;
       end;
       ^Y:begin Data^ := ''; CurPos := 0 end;
       else Exit;
      end
     end;
     if ExtendBlock
      then AdjustSelectBlock
      else begin SelStart := CurPos; SelEnd := CurPos end;
     if FirstPos > CurPos then FirstPos := CurPos;
     I := CurPos - Size.X + 2;
     if FirstPos < I then FirstPos := I;
     DrawView; ClearEvent(Event);
    end;
   end;
  end;
 end;

procedure TInputLine.SelectAll(Enable: Boolean);
 begin
  CurPos := 0; FirstPos := 0; SelStart := 0;
  if Enable then SelEnd := Length(Data^) else SelEnd := 0;
  DrawView;
 end;

procedure TInputLine.SetData(var Rec);
 begin
  Move(Rec, Data^[0], DataSize);
  SelectAll(True);
 end;

procedure TInputLine.SetState(AState: Word; Enable: Boolean);
 begin
  TView.SetState(AState, Enable);
  if (AState = sfSelected) or ((AState = sfActive) and
     (State and sfSelected <> 0))
   then SelectAll(Enable)
   else if AState = sfFocused then DrawView;
 end;

function TInputLine.Valid(Command: Word): Boolean;
 begin Valid := inherited Valid(Command) end;

{ TButton }

constructor TButton.Init(var Bounds: TRect; ATitle: TTitleStr;
  ACommand: Word; AFlags: Word);
 begin
  TView.Init(Bounds);
  Options := Options or
   (ofSelectable + ofFirstClick + ofPreProcess + ofPostProcess);
  EventMask := EventMask or evBroadcast;
  if not CommandEnabled(ACommand) then State := State or sfDisabled;
  Flags := AFlags;
  AmDefault:=AFlags and bfDefault <> 0;
  Title := NewStr(ATitle);
  Command := ACommand;
 end;

destructor TButton.Done;
 begin DisposeStr(Title); TView.Done end;

procedure TButton.Draw;
 begin DrawState(False) end;

procedure TButton.DrawState(Down: Boolean);
 var
  CButton, CShadow: Word;
  Ch: Char;
  I, S, Y, T: Integer;
  B: TDrawBuffer;

 procedure DrawTitle;
  var L, SCOff: Integer;
  begin
   if Flags and bfLeftJust <> 0 then L := 1 else begin
    L := (S - CStrLen(Title^) - 1) div 2;
    if L < 1 then L := 1;
   end;
   MoveCStr(B[I + L], Title^, CButton);
   if ShowMarkers and not Down then begin
    if State and sfSelected <> 0 then SCOff := 0 else
      if AmDefault then SCOff := 2 else SCOff := 4;
    WordRec(B[0]).Lo := Byte(SpecialChars[SCOff]);
    WordRec(B[S]).Lo := Byte(SpecialChars[SCOff + 1]);
   end;
  end;

 begin
  if State and sfDisabled <> 0 then CButton := GetColor($0404) else
  begin
    CButton := GetColor($0501);
    if State and sfActive <> 0 then
      if State and sfSelected <> 0 then CButton := GetColor($0703) else
        if AmDefault then CButton := GetColor($0602);
  end;
  CShadow := GetColor(8);
  S := Size.X - 1;
  T := Size.Y div 2 - 1;
  for Y := 0 to Size.Y - 2 do
  begin
    MoveChar(B, ' ', Byte(CButton), Size.X);
    WordRec(B[0]).Hi := CShadow;
    if Down then
    begin
      WordRec(B[1]).Hi := CShadow;
      Ch := ' ';
      I := 2;
    end else
    begin
      WordRec(B[S]).Hi := Byte(CShadow);
      if ShowMarkers then Ch := ' ' else
      begin
        if Y = 0 then
          WordRec(B[S]).Lo := Byte('') else
          WordRec(B[S]).Lo := Byte('');
        Ch := '';
      end;
      I := 1;
    end;
    if (Y = T) and (Title <> nil) then DrawTitle;
    if ShowMarkers and not Down then
    begin
      WordRec(B[1]).Lo := Byte('[');
      WordRec(B[S - 1]).Lo := Byte(']');
    end;
    WriteLine(0, Y, Size.X, 1, B);
  end;
  MoveChar(B[0], ' ', Byte(CShadow), 2);
  MoveChar(B[2], Ch, Byte(CShadow), S - 1);
  WriteLine(0, Size.Y - 1, Size.X, 1, B);
 end;

function TButton.GetPalette: PPalette;
 const P: String[Length(CButton)] = CButton;
 begin GetPalette := @P end;

procedure TButton.HandleEvent(var Event: TEvent);
 var Down: Boolean; C: Char; Mouse: TPoint; ClickRect: TRect;
 begin
  GetExtent(ClickRect);
  Inc(ClickRect.A.X);
  Dec(ClickRect.B.X);
  Dec(ClickRect.B.Y);
  if Event.What = evMouseDown then begin
    MakeLocal(Event.Where, Mouse);
    if not ClickRect.Contains(Mouse) then ClearEvent(Event);
  end;
  if Flags and bfGrabFocus <> 0 then TView.HandleEvent(Event);
  case Event.What of
   evMouseDown:begin
    if State and sfDisabled = 0 then begin
     Inc(ClickRect.B.X); Down := False;
     repeat
      MakeLocal(Event.Where, Mouse);
      if Down <> ClickRect.Contains(Mouse)
       then begin Down := not Down; DrawState(Down) end;
     until not MouseEvent(Event, evMouseMove);
     if Down then begin Press; DrawState(False) end;
    end;
    ClearEvent(Event);
   end;
   evKeyDown:begin
    C := HotKey(Title^);
    if (Event.KeyCode = GetAltCode(C)) or
       (Owner^.Phase = phPostProcess) and (C <> #0) and
       (Upcase(Event.CharCode) = C) or
       (State and sfFocused <> 0) and (Event.CharCode = ' ')
     then begin Press; ClearEvent(Event) end;
   end;
   evBroadcast:case Event.Command of
    cmDefault:
     if AmDefault then begin Press; ClearEvent(Event) end;
    cmGrabDefault, cmReleaseDefault:
     if Flags and bfDefault <> 0 then begin
      AmDefault := Event.Command = cmReleaseDefault;
      DrawView;
     end;
    cmCommandSetChanged:begin
     SetState(sfDisabled, not CommandEnabled(Command));
     DrawView;
    end;
   end;
  end;
 end;

procedure TButton.MakeDefault(Enable: Boolean);
 var C: Word;
 begin if Flags and bfDefault = 0 then begin
  if Enable then C := cmGrabDefault else C := cmReleaseDefault;
  Message(Owner, evBroadcast, C, @Self);
  AmDefault := Enable; DrawView;
 end end;

procedure TButton.Press;
 var E: TEvent;
 begin
  Message(Owner, evBroadcast, cmRecordHistory, nil);
  if Flags and bfBroadcast <> 0
   then Message(Owner, evBroadcast, Command, @Self)
   else begin
    E.What := evCommand;
    E.Command := Command;
    E.InfoPtr := @Self;
    PutEvent(E);
   end;
 end;

procedure TButton.SetState(AState: Word; Enable: Boolean);
 begin
  TView.SetState(AState, Enable);
  if AState and (sfSelected + sfActive) <> 0 then DrawView;
  if AState and sfFocused <> 0 then MakeDefault(Enable);
 end;

{ TCheckBox }

constructor TCheckBox.Init(var Bounds: TRect);
 begin
  TView.Init(Bounds);
  Options := Options or
   (ofSelectable + ofFirstClick + ofPreProcess + ofPostProcess);
  SetCursor(2,0); ShowCursor;
 end;

destructor TCheckBox.Done;
 begin TView.Done end;

function TCheckBox.DataSize: Word;
 begin DataSize := SizeOf(boolean) end;

procedure TCheckBox.GetData(var Rec);
 begin boolean(Rec) := Value end;

function TCheckBox.GetHelpCtx: Word;
 begin
{
  if HelpCtx = hcNoContext
   then GetHelpCtx := hcNoContext
   else
}
   GetHelpCtx := HelpCtx;
 end;

function TCheckBox.GetPalette: PPalette;
 const P: String[Length(CCluster)] = CCluster;
 begin GetPalette := @P end;

procedure TCheckBox.HandleEvent(var Event: TEvent);
 var
  Mouse: TPoint;
 begin
  TView.HandleEvent(Event);
  if (Options and ofSelectable) = 0 then Exit;
  if Event.What = evMouseDown then
  begin
    MakeLocal(Event.Where, Mouse);
    DrawView;
    repeat
      MakeLocal(Event.Where, Mouse);
      if FindSel(Mouse) then ShowCursor else HideCursor;
    until not MouseEvent(Event,evMouseMove); {Wait for mouse up}
    ShowCursor;
    MakeLocal(Event.Where, Mouse);
    if FindSel(Mouse) then begin Press; DrawView end;
    ClearEvent(Event);
  end else if Event.What = evKeyDown then
  begin
    if (Event.CharCode = ' ') and (State and sfFocused <> 0) then begin
     Press; DrawView;
     ClearEvent(Event);
    end;
  end;
 end;

procedure TCheckBox.SetData(var Rec);
 begin Value := boolean(Rec); DrawView end;

procedure TCheckBox.SetState(AState: Word; Enable: Boolean);
 begin
  TView.SetState(AState, Enable);
  if AState = sfFocused then DrawView;
 end;

function TCheckBox.Mark: Boolean;
 begin Mark := Value end;

{
function TCheckBox.MultiMark: Byte;
 begin MultiMark := Byte(Mark = True) end;
}

procedure TCheckBox.Press;
 begin Value := not Value end;

function TCheckBox.FindSel(P: TPoint): boolean;
 var R: TRect;
 begin GetExtent(R); FindSel:=R.Contains(P) end;

procedure TCheckBox.Draw { MultiBox(const Icon, Marker: String)} ;
 const Icon = ' [ ] ';
 var CNorm, CSel, CDis, Color: Word; B: TDrawBuffer;
 begin
  CNorm := GetColor($0301);
  CSel := GetColor($0402);
  CDis := GetColor($0505);
  MoveChar(B, ' ', Byte(CNorm), Size.X);
  if State and sfFocused <> 0 then Color := CSel else Color := CNorm;
  MoveChar(B[0], ' ', Byte(Color), Size.X);
  MoveStr(B[0], Icon, Byte(Color));
  if Mark
   then WordRec(B[2]).Lo := Byte('X')
   else WordRec(B[2]).Lo := Byte(' ');
  { Marker[MultiMark + 1]); }
  WriteBuf(0, 0, Size.X, 1, B);
  SetCursor(2,0);
 end;

{
procedure TCheckBox.Draw;
 begin
  DrawMultiBox(Button, ' X');
 end;
}

{ TStaticText }

constructor TStaticText.Init(var Bounds: TRect; const AText: String);
 begin TView.Init(Bounds); Text := NewStr(AText) end;

destructor TStaticText.Done;
 begin DisposeStr(Text); TView.Done end;

procedure TStaticText.Draw;
 var
  Color: Byte; Center: Boolean; I, J, L, P, Y: Integer;
  B: TDrawBuffer; S: String;
 begin
  Color := GetColor(1); GetText(S);
  L := Length(S); P := 1; Y := 0;
  Center := False;
  while Y < Size.Y do begin
   MoveChar(B, ' ', Color, Size.X);
   if P <= L then begin
    if S[P] = #3 then begin Center := True; Inc(P) end;
    I := P;
    repeat
     J := P;
     while (P <= L) and (S[P] = ' ') do Inc(P);
     while (P <= L) and (S[P] <> ' ') and (S[P] <> #13) do Inc(P);
    until (P > L) or (P >= I + Size.X) or (S[P] = #13);
    if P > I + Size.X then if J > I then P := J else P := I + Size.X;
    if Center then J := (Size.X - P + I) div 2 else J := 0;
    MoveBuf(B[J], S[I], Color, P - I);
    while (P <= L) and (S[P] = ' ') do Inc(P);
    if (P <= L) and (S[P] = #13) then begin
     Center := False; Inc(P);
     if (P <= L) and (S[P] = #10) then Inc(P);
    end;
   end;
   WriteLine(0, Y, Size.X, 1, B);
   Inc(Y);
  end;
 end;

function TStaticText.GetPalette: PPalette;
 const P: String[Length(CStaticText)] = CStaticText;
 begin GetPalette := @P end;

procedure TStaticText.GetText(var S: String);
 begin if Text <> nil then S := Text^ else S := '' end;

{ TParamText }

constructor TParamText.Init(var Bounds: TRect; const AText: String;
  AParamCount: Integer);
 begin
  TStaticText.Init(Bounds, AText);
  ParamCount := AParamCount;
 end;

function TParamText.DataSize: Word;
 begin DataSize := ParamCount * SizeOf(Longint) end;

procedure TParamText.GetText(var S: String);
 begin
  if Text <> nil
   then FormatStr(S, Text^, ParamList^) else S := '';
 end;

procedure TParamText.SetData(var Rec);
 begin ParamList := @Rec; DrawView end;

{ TLabel }

constructor TLabel.Init(var Bounds: TRect; const AText: String; ALink: PView);
 begin
  TStaticText.Init(Bounds, AText);
  Link := ALink;
  Options := Options or (ofPreProcess + ofPostProcess);
  EventMask := EventMask or evBroadcast;
 end;

procedure TLabel.Draw;
 var Color: Word; B: TDrawBuffer; SCOff: Byte;
 begin
  if Light
   then begin Color := GetColor($0402); SCOff := 0 end
   else begin Color := GetColor($0301); SCOff := 4 end;
  MoveChar(B[0], ' ', Byte(Color), Size.X);
  if Text <> nil then MoveCStr(B[1], Text^, Color);
  if ShowMarkers then WordRec(B[0]).Lo := Byte(SpecialChars[SCOff]);
  WriteLine(0, 0, Size.X, 1, B);
 end;

function TLabel.GetPalette: PPalette;
 const P: String[Length(CLabel)] = CLabel;
 begin GetPalette := @P end;

procedure TLabel.HandleEvent(var Event: TEvent);
 var C: Char;
 procedure FocusLink;
  begin
   if (Link <> nil) and (Link^.Options and ofSelectable <> 0)
    then Link^.Focus;
   ClearEvent(Event);
  end;
 begin
  inherited HandleEvent(Event);
  if Event.What = evMouseDown
   then FocusLink
   else if Event.What = evKeyDown then begin
    C := HotKey(Text^);
    if (GetAltCode(C) = Event.KeyCode) or
       ((C <> #0) and (Owner^.Phase = phPostProcess) and
        (UpCase(Event.CharCode) = C)) then FocusLink
   end else if Event.What = evBroadcast then
    if ((Event.Command = cmReceivedFocus) or
       (Event.Command = cmReleasedFocus)) and
       (Link <> nil)
     then begin
      Light := Link^.State and sfFocused <> 0;
      DrawView;
    end;
 end;

end.
