unit Drivers;

{$I define.pp}

interface

uses Trace, Objects;

const

{ Event codes }

  evMouseDown = $0001;
  evMouseUp   = $0002;
  evMouseMove = $0004;
  evMouseAuto = $0008;
  evKeyDown   = $0010;
  evCommand   = $0100;
  evBroadcast = $0200;

{ Event masks }

  evNothing   = $0000;
  evMouse     = $000F;
  evKeyboard  = $0010;
  evMessage   = $FF00;

{ Extended key codes }

  kbEsc       = $011B;  kbAltSpace  = $0200;  kbCtrlIns   = $0400;
  kbShiftIns  = $0500;  kbCtrlDel   = $0600;  kbShiftDel  = $0700;
  kbBack      = $0E08;  kbCtrlBack  = $0E7F;  kbShiftTab  = $0F00;
  kbTab       = $0F09;  kbAltQ      = $1000;  kbAltW      = $1100;
  kbAltE      = $1200;  kbAltR      = $1300;  kbAltT      = $1400;
  kbAltY      = $1500;  kbAltU      = $1600;  kbAltI      = $1700;
  kbAltO      = $1800;  kbAltP      = $1900;  kbCtrlEnter = $1C0A;
  kbEnter     = $1C0D;  kbAltA      = $1E00;  kbAltS      = $1F00;
  kbAltD      = $2000;  kbAltF      = $2100;  kbAltG      = $2200;
  kbAltH      = $2300;  kbAltJ      = $2400;  kbAltK      = $2500;
  kbAltL      = $2600;  kbAltZ      = $2C00;  kbAltX      = $2D00;
  kbAltC      = $2E00;  kbAltV      = $2F00;  kbAltB      = $3000;
  kbAltN      = $3100;  kbAltM      = $3200;  kbF1        = $3B00;
  kbF2        = $3C00;  kbF3        = $3D00;  kbF4        = $3E00;
  kbF5        = $3F00;  kbF6        = $4000;  kbF7        = $4100;
  kbF8        = $4200;  kbF9        = $4300;  kbF10       = $4400;
  kbHome      = $4700;  kbUp        = $4800;  kbPgUp      = $4900;
  kbGrayMinus = $4A2D;  kbLeft      = $4B00;  kbRight     = $4D00;
  kbGrayPlus  = $4E2B;  kbEnd       = $4F00;  kbDown      = $5000;
  kbPgDn      = $5100;  kbIns       = $5200;  kbDel       = $5300;
  kbShiftF1   = $5400;  kbShiftF2   = $5500;  kbShiftF3   = $5600;
  kbShiftF4   = $5700;  kbShiftF5   = $5800;  kbShiftF6   = $5900;
  kbShiftF7   = $5A00;  kbShiftF8   = $5B00;  kbShiftF9   = $5C00;
  kbShiftF10  = $5D00;  kbCtrlF1    = $5E00;  kbCtrlF2    = $5F00;
  kbCtrlF3    = $6000;  kbCtrlF4    = $6100;  kbCtrlF5    = $6200;
  kbCtrlF6    = $6300;  kbCtrlF7    = $6400;  kbCtrlF8    = $6500;
  kbCtrlF9    = $6600;  kbCtrlF10   = $6700;  kbAltF1     = $6800;
  kbAltF2     = $6900;  kbAltF3     = $6A00;  kbAltF4     = $6B00;
  kbAltF5     = $6C00;  kbAltF6     = $6D00;  kbAltF7     = $6E00;
  kbAltF8     = $6F00;  kbAltF9     = $7000;  kbAltF10    = $7100;
  kbCtrlPrtSc = $7200;  kbCtrlLeft  = $7300;  kbCtrlRight = $7400;
  kbCtrlEnd   = $7500;  kbCtrlPgDn  = $7600;  kbCtrlHome  = $7700;
  kbAlt1      = $7800;  kbAlt2      = $7900;  kbAlt3      = $7A00;
  kbAlt4      = $7B00;  kbAlt5      = $7C00;  kbAlt6      = $7D00;
  kbAlt7      = $7E00;  kbAlt8      = $7F00;  kbAlt9      = $8000;
  kbAlt0      = $8100;  kbAltMinus  = $8200;  kbAltEqual  = $8300;
  kbCtrlPgUp  = $8400;  kbAltBack   = $0800;  kbNoKey     = $0000;

{ Keyboard state and shift masks }

  kbRightShift  = $0001;
  kbLeftShift   = $0002;
  kbCtrlShift   = $0004;
  kbAltShift    = $0008;
  kbScrollState = $0010;
  kbNumState    = $0020;
  kbCapsState   = $0040;
  kbInsState    = $0080;

{ Mouse button state masks }

  mbLeftButton  = $01;
  mbRightButton = $02;

type

{ Event record }

  PEvent = ^TEvent;
  TEvent = record
    What: Word;
    case Word of
      evNothing: ();
      evMouse: (
        Buttons: Byte;
        Double: Boolean;
        Where: TPoint);
      evKeyDown: (
        case Integer of
	  0: (KeyCode: Word);
          1: (CharCode: Char;
              ScanCode: Byte));
      evMessage: (
        Command: Word;
        case Word of
          0: (InfoPtr: Pointer);
          1: (InfoLong: Longint);
          2: (InfoWord: Word);
          3: (InfoInt: Integer);
          4: (InfoByte: Byte);
          5: (InfoChar: Char));
  end;

const

{ Initialized variables }

  ButtonCount: Byte = 0;
  MouseEvents: Boolean = False;
  MouseReverse: Boolean = False;
  DoubleDelay: Word = 8;
  RepeatDelay: Word = 8;

var

{ Uninitialized variables }

  MouseIntFlag: Byte;
  MouseButtons: Byte;
  MouseWhere: TPoint;

{ Event manager routines }

procedure InitEvents;
procedure DoneEvents;
procedure ShowMouse;
procedure HideMouse;
procedure GetMouseEvent(var Event: TEvent);
procedure GetKeyEvent(var Event: TEvent);
function GetShiftState: Byte;

{ ******** SCREEN MANAGER ******** }

const

{ Screen modes }

  smBW80    = $0002;
  smCO80    = $0003;
  smMono    = $0007;
  smFont8x8 = $0100;

const

{ Initialized variables }

  StartupMode: Word = $FFFF;

{ Screen manager routines }

procedure InitVideo;
procedure DoneVideo;
procedure SetVideoMode(Mode: Word);

const

  SysColorAttr: Word = $4E4F;
  SysMonoAttr: Word = $7070;

{ ******** UTILITY ROUTINES ******** }

{ Keyboard support routines }

function GetAltChar(KeyCode: Word): Char;
function GetAltCode(Ch: Char): Word;
function GetCtrlChar(KeyCode: Word): Char;
function GetCtrlCode(Ch: Char): Word;
function CtrlToArrow(KeyCode: Word): Word;

{ String routines }

procedure FormatStr(var Result: String; const Format: String; var Params);

{ Buffer move routines }

procedure MoveBuf(var Dest; var Source; Attr: Byte; Count: Word);
procedure MoveChar(var Dest; C: Char; Attr: Byte; Count: Word);
procedure MoveCStr(var Dest; const Str: String; Attrs: Word);
procedure MoveStr(var Dest; const Str: String; Attr: Byte);
function CStrLen(const S: String): Integer;

implementation

uses crt, Screen;

procedure DetectMouse;
 begin ButtonCount:=0 end;
procedure InitEvents; begin end;
procedure DoneEvents; begin end;
procedure ShowMouse;
 begin if ButtonCount>0 then begin
 end end;
procedure HideMouse;
 begin if ButtonCount>0 then begin
 end end;
procedure GetMouseEvent(var Event: TEvent);
 begin
  Event.What:=evNothing;
 end;
procedure SetMouseRange;
 begin
 end;

const ShiftState:byte = 0;

var MyChar:char; MyScan, InsState:byte;

procedure CrtGetKey;
 function EvalShift1(b:byte):byte;
  begin
   EvalShift1:=$00;
   if b<$34
    then if b=$0F
     then EvalShift1:=$02 { Shift-TAB }
     else EvalShift1:=$08
    else case b of
     $68..$71,$78..$83:EvalShift1:=$08;
     $5E..$67,$72..$77:EvalShift1:=$04;
     $54..$5D         :EvalShift1:=$02;
    end;
  end;
 function EvalShift(b:byte):byte;
  const
   DShift:array[0..31] of byte = (
    $00, $02, $02, $02, $02, $02, $02, $00,
    $02, $02, $02, $02, $00, $00, $00, $00,
    $00, $00, $00, $00, $00, $00, $00, $00,
    $00, $00, $02, $00, $02, $00, $02, $02);
  begin
   EvalShift:=$00; { no specials }
   case b shr 5 of
    0:case b of
       $09,$0D,$1B:; { TAB, CR, esc }
       else EvalShift:=$04; { Ctrl }
      end;
    1:EvalShift:=DShift[b and $1F];
    2,4,6:EvalShift:=$02; { RigthShift }
    3,5,7:;
   end;
   case b of
    $5B..$5D,$7B..$7E:EvalShift:=(b and $20) shr 4;
   end
  end;
 function EvalScan(b:byte):byte;
  const
   DScan:array[0..31] of byte = (
    $39, $02, $28, $04, $05, $06, $08, $28,
    $0A, $0B, $09, $0D, $33, $0C, $34, $35,
    $0B, $02, $03, $04, $05, $06, $07, $08,
    $09, $0A, $27, $27, $33, $0D, $34, $35);
   LScan:array[0..31] of byte = (
    $29, $1E, $30, $2E, $20, $12, $21, $22,
    $23, $17, $24, $25, $26, $32, $31, $18,
    $19, $10, $13, $1F, $14, $16, $2F, $11,
    $2D, $15, $2C, $1A, $2B, $1B, $29, $0C);
  begin
   if (b and $E0)=$20  { digits / leters }
    then EvalScan:=DScan[b and $1F]
    else case b of
     $09:EvalScan:=$0F; { TAB }
     $0D:EvalScan:=$1C; { CR }
     $1B:EvalScan:=$01; { esc }
     $40:EvalScan:=$03; { @ }
     $5E:EvalScan:=$07; { ^ }
     $60:EvalScan:=$29; { ` }
     else EvalScan:=LScan[b and $1F];
    end;
  end;
 function EvalScanZ(b:byte):byte;
  begin
   EvalScanZ:=b;
   if b in [$3B..$44] { F1..F10 -> Alt-F1..Alt-F10}
    then EvalScanZ:=b+$2D;
  end;
 begin
  MyChar:=Readkey; MyScan:=ord(MyChar); ShiftState:=0;

  if MyChar=chr($1A) then begin { ^Z - replace Alt for Linux OS }
   MyChar:=ReadKey; MyScan:=ord(MyChar);
   if MyScan=0 then begin
    MyScan:=EvalScanZ(ord(ReadKey));
   end else begin
    MyScan:=EvalScan(ord(MyChar));
    if MyScan in [$02..$0D] then inc(MyScan,$76);
    MyChar:=chr(0);
   end;
   ShiftState:=$08; { emulate Alt }
  end else

  if MyScan=0 then begin
   MyScan:=ord(ReadKey);
   if MyScan=$52 then InsState:=InsState xor $80;
   ShiftState:=EvalShift1(MyScan);
  end else begin
   MyScan:=EvalScan(ord(MyChar));
   ShiftState:=EvalShift(ord(MyChar)) or InsState;
  end;

 end;

procedure GetKeyEvent(var Event: TEvent);
 begin
  if keypressed then begin
   CrtGetKey;
   Event.What:=evKeyDown;
   Event.CharCode:=MyChar;
   Event.ScanCode:=MyScan;
   { special handling for kbEnter }
   if Event.KeyCode=$240A then Event.KeyCode:=$1C0D;
  end else Event.What:=evNothing;
 end;

function GetShiftState: Byte;
 begin GetShiftState:=ShiftState end;

{ ******** SCREEN MANAGER ******** }

{ Set CRT data areas and mouse range }

procedure SetCrtData;
 begin
  MyScreen.ScreenMode:=smCO80;
  if ButtonCount<>0 then SetMouseRange;
 end;

procedure DetectVideo;
 begin MyScreen.ScreenMode:=smCO80 end;

procedure InitVideo;
 begin
  writeln(chr(27)+'(K');
  MyScreen.Init;
  SetCrtData;
 end;

procedure DoneVideo;
 begin
  MyScreen.Done;
 end;

procedure SetVideoMode(Mode: Word);
 begin
 end;

{ Keyboard support routines }

const

  AltCodes1: array[$10..$32] of Char =
    'QWERTYUIOP'#0#0#0#0'ASDFGHJKL'#0#0#0#0#0'ZXCVBNM';

  AltCodes2: array[$78..$83] of Char =
    '1234567890-=';

function GetAltChar(KeyCode: Word): Char;
 begin
  GetAltChar := #0;
  if Lo(KeyCode) = 0 then case Hi(KeyCode) of
   $02: GetAltChar := #240;
   $10..$32: GetAltChar := AltCodes1[Hi(KeyCode)];
   $78..$83: GetAltChar := AltCodes2[Hi(KeyCode)];
  end;
 end;

function GetAltCode(Ch: Char): Word;
 var I: Word;
 begin
  GetAltCode := 0;
  if Ch = #0 then Exit;
  Ch := UpCase(Ch);
  if Ch = #240 then begin GetAltCode := $0200; Exit end;
  for I := $10 to $32 do if AltCodes1[I] = Ch
   then begin GetAltCode := I shl 8; Exit end;
  for I := $78 to $83 do if AltCodes2[I] = Ch
   then begin GetAltCode := I shl 8; Exit end;
 end;

function GetCtrlChar(KeyCode: Word): Char;
 begin
  GetCtrlChar := #0;
  if (Lo(KeyCode) <> 0) and (Lo(KeyCode) <= Byte('Z') - Byte('A') + 1) then
    GetCtrlChar := Char(Lo(KeyCode) + Byte('A') - 1);
 end;

function GetCtrlCode(Ch: Char): Word;
 begin
  GetCtrlCode := GetAltCode(Ch) or (Byte(UpCase(Ch)) - Byte('A') + 1);
 end;

function CtrlToArrow(KeyCode: Word): Word;
 const
  NumCodes = 11;
  CtrlCodes: array[0..NumCodes-1] of Char = ^S^D^E^X^A^F^G^V^R^C^H;
  ArrowCodes: array[0..NumCodes-1] of Word =
    (kbLeft, kbRight, kbUp, kbDown, kbHome, kbEnd, kbDel, kbIns,
     kbPgUp, kbPgDn, kbBack);
 var I: Integer;
 begin
  CtrlToArrow := KeyCode;
  for I := 0 to NumCodes - 1 do
   if WordRec(KeyCode).Lo = Byte(CtrlCodes[I])
    then begin CtrlToArrow := ArrowCodes[I]; Exit end;
 end;

{ String formatting routines }

FUNCTION CStrLen (Const S: String): Integer;
 VAR I, J: Integer;
 BEGIN
  J := 0;                                            { Set result to zero }
  For I := 1 To Length(S) Do If (S[I] <> '~') Then Inc(J); { Inc count if not ~ }
  CStrLen := J;                                      { Return length }
 END;

PROCEDURE MoveStr (Var Dest; Const Str: String; Attr: Byte);
 VAR I: Word;
 BEGIN
   For I := 1 To Length(Str) Do Begin                 { For each character }
     If (Attr<>0) Then
       WordRec(TWordArray(Dest)[I-1]).Hi := Attr;     { Copy attribute }
     WordRec(TWordArray(Dest)[I-1]).Lo:=Byte(Str[I]); { Copy string char }
   End;
 END;

PROCEDURE MoveCStr (Var Dest; Const Str: String; Attrs: Word);
 VAR B: Byte; I, J: Word;
 BEGIN
   J := 0;                                            { Start position }
   For I := 1 To Length(Str) Do Begin                 { For each character }
     If (Str[I] <> '~') Then Begin                    { Not tilde character }
       If (Lo(Attrs)<>0) Then
         WordRec(TWordArray(Dest)[J]).Hi := Lo(Attrs);{ Copy attribute }
       WordRec(TWordArray(Dest)[J]).Lo:=Byte(Str[I]); { Copy string char }
       Inc(J);                                        { Next position }
     End Else Begin
       B := Hi(Attrs);                                { Hold attribute }
       WordRec(Attrs).Hi := Lo(Attrs);                { Copy low to high }
       WordRec(Attrs).Lo := B;                        { Complete exchange }
     End;
   End;
 END;

PROCEDURE MoveBuf (Var Dest, Source; Attr: Byte; Count: Word);
 VAR I: Word;
 BEGIN
   For I := 1 To Count Do Begin
     If (Attr<>0) Then
       WordRec(TWordArray(Dest)[I-1]).Hi := Attr;     { Copy attribute }
     WordRec(TWordArray(Dest)[I-1]).Lo :=
      TByteArray(Source)[I-1];                        { Copy source data }
   End;
 END;

PROCEDURE MoveChar (Var Dest; C: Char; Attr: Byte; Count: Word);
 VAR I: Word;
 BEGIN
   For I := 1 To Count Do Begin
     If (Attr<>0) Then
       WordRec(TWordArray(Dest)[I-1]).Hi := Attr;     { Copy attribute }
     If (Ord(C)<>0) Then
       WordRec(TWordArray(Dest)[I-1]).Lo := Byte(C);  { Copy character }
   End;
 END;

{-------------------------------------------------------------------------}

{ Functions below are not fully completed yet !!!! }

procedure FormatStr(var Result: String; const Format: String; var Params);
 function HexByte(b:byte):string;
  const Hex:string = '0123456789ABCDEF';
  begin HexByte:=Hex[b div 16 + 1] + Hex[b mod 16 + 1] end;
 function HexLong(l:longint; n:integer):string;
  var w:string[20];
  begin
   w:=
    HexByte((l shr 24) and $ff)+
    HexByte((l shr 16) and $ff)+
    HexByte((l shr  8) and $ff)+
    HexByte(l and $ff);
   if (n>0) and (n<8) then w:=copy(w,9-n,n);
   HexLong:='$'+w;
  end;
 type
  PtrArr = array[0..99] of PString;
  IntArr = array[0..99] of longint;
 var i,PCnt,State:byte; w:char; Just:boolean; ws:string; nnn:integer;
 begin
  Result:='';
  i:=1; State:=0; PCnt:=0;
  while i<=Length(Format) do begin
   w:=Format[i];
   case State of
    0:if w='%'
       then begin State:=1; nnn:=0; Just:=True end
       else Result:=Result+w;
    1:case w of
       '-':Just:=False;
       '0'..'9':nnn:=10*nnn+ord(w)-ord('0');
       's','d','c','x','#':begin
        case w of
         's':begin
              ws:=PtrArr(Params)[PCnt]^;
              if nnn>0 then begin
               while length(ws)<nnn do if Just
                then ws:=' '+ws else ws:=ws+' ';
               ws:=copy(ws,1,nnn);
              end;
              Result:=Result+ws
             end;
         'c':Result:=Result+chr(IntArr(Params)[PCnt] and $ff);
         'd':begin
              if nnn=0
               then Str(IntArr(Params)[PCnt],ws)
               else Str(IntArr(Params)[PCnt]:nnn,ws);
              Result:=Result+ws;
             end;
         'x':Result:=Result+HexLong(IntArr(Params)[PCnt],nnn);
         '#':PCnt:=nnn-1;
        end;
        State:=0; inc(PCnt);
       end;
      end;
   end;
   inc(i);
  end;
 end;

{ Drivers unit initialization and shutdown }

var SaveExit: Pointer;

procedure ExitDrivers;
 begin
  DoneEvents;
  ExitProc := SaveExit;
 end;

begin
 DetectMouse;
 DetectVideo;
 SaveExit := ExitProc;
 ExitProc := @ExitDrivers;
end.
