unit Views;

{$I define.pp}

interface

uses Objects, Drivers{, Trace};

const

{ TView State masks }

  sfVisible     = $0001;
  sfCursorVis   = $0002;
  sfCursorIns   = $0004;
  sfShadow      = $0008;
  sfActive      = $0010;
  sfSelected    = $0020;
  sfFocused     = $0040;
  sfDragging    = $0080;
  sfDisabled    = $0100;
  sfModal       = $0200;
  sfDefault     = $0400;
  sfExposed     = $0800;

{ TView Option masks }

  ofSelectable  = $0001;
  ofTopSelect   = $0002;
  ofFirstClick  = $0004;
  ofFramed      = $0008;
  ofPreProcess  = $0010;
  ofPostProcess = $0020;
  ofBuffered    = $0040;
  ofTileable    = $0080;
  ofCenterX     = $0100;
  ofCenterY     = $0200;
  ofCentered    = $0300;
  ofValidate    = $0400;
  ofVersion     = $3000;
  ofVersion10   = $0000;
  ofVersion20   = $1000;

{ TView GrowMode masks }

  gfGrowLoX = $01;
  gfGrowLoY = $02;
  gfGrowHiX = $04;
  gfGrowHiY = $08;
  gfGrowAll = $0F;
  gfGrowRel = $10;

{ TView DragMode masks }

  dmDragMove = $01;
  dmDragGrow = $02;
  dmLimitLoX = $10;
  dmLimitLoY = $20;
  dmLimitHiX = $40;
  dmLimitHiY = $80;
  dmLimitAll = $F0;

{ TView Help context codes }

  hcNoContext = 0;
  hcDragging  = 1;

{ TScrollBar part codes }

  sbLeftArrow  = 0;
  sbRightArrow = 1;
  sbPageLeft   = 2;
  sbPageRight  = 3;
  sbUpArrow    = 4;
  sbDownArrow  = 5;
  sbPageUp     = 6;
  sbPageDown   = 7;
  sbIndicator  = 8;

{ TScrollBar options for TWindow.StandardScrollBar }

  sbHorizontal     = $0000;
  sbVertical       = $0001;
  sbHandleKeyboard = $0002;

{ TWindow Flags masks }

  wfMove  = $01;
  wfGrow  = $02;
  wfClose = $04;
  wfZoom  = $08;

{ TWindow number constants }

  wnNoNumber = 0;

{ TWindow palette entries }

  wpBlueWindow = 0;
  wpCyanWindow = 1;
  wpGrayWindow = 2;

{ Standard command codes }

  cmValid   = 0;
  cmQuit    = 1;
  cmError   = 2;
  cmMenu    = 3;
  cmClose   = 4;
  cmZoom    = 5;
  cmResize  = 6;
  cmNext    = 7;
  cmPrev    = 8;
  cmHelp    = 9;

{ Application command codes }

  cmCut     = 20;
  cmCopy    = 21;
  cmPaste   = 22;
  cmUndo    = 23;
  cmClear   = 24;
  cmTile    = 25;
  cmCascade = 26;

{ TDialog standard commands }

  cmOK      = 10;
  cmCancel  = 11;
  cmYes     = 12;
  cmNo      = 13;
  cmDefault = 14;

{ Standard messages }

  cmReceivedFocus     = 50;
  cmReleasedFocus     = 51;
  cmCommandSetChanged = 52;

{ TScrollBar messages }

  cmScrollBarChanged  = 53;
  cmScrollBarClicked  = 54;

{ TWindow select messages }

  cmSelectWindowNum   = 55;

{ Color palettes }

  CFrame      = #1#1#2#2#3;
  CScrollBar  = #4#5#5;
  CScroller   = #6#7;

  CBlueWindow = #8#9#10#11#12#13#14#15;
  CCyanWindow = #16#17#18#19#20#21#22#23;
  CGrayWindow = #24#25#26#27#28#29#30#31;

{ TDrawBuffer maximum view width }

  MaxViewWidth = 132;

type

{ Command sets }

  PCommandSet = ^TCommandSet;
  TCommandSet = set of byte;

{ Color palette type }

  PPalette = ^TPalette;
  TPalette = String;

{ TDrawBuffer, buffer used by draw methods }

  TDrawBuffer = array[0..MaxViewWidth - 1] of Word;

  PView = ^TView;
  PGroup = ^TGroup;

{ TView object }

  TView = object(TObject)
    Owner: PGroup;
    Next: PView;
    Origin: TPoint;
    Size: TPoint;
    Cursor: TPoint;
    GrowMode: Byte;
    DragMode: Byte;
    HelpCtx: Word;
    State: Word;
    Options: Word;
    EventMask: Word;
    constructor Init(var Bounds: TRect);
    destructor Done; virtual;
    procedure Awaken; virtual;
    procedure BlockCursor;
    procedure CalcBounds(var Bounds: TRect; Delta: TPoint); virtual;
    procedure ChangeBounds(var Bounds: TRect); virtual;
    procedure ClearEvent(var Event: TEvent);
    function CommandEnabled(Command: Word): Boolean;
    function DataSize: Word; virtual;
    procedure DisableCommands(Commands: TCommandSet);
    procedure DragView(Event: TEvent; Mode: Byte;
      var Limits: TRect; MinSize, MaxSize: TPoint);
    procedure Draw; virtual;
    procedure DrawView;
    procedure EnableCommands(Commands: TCommandSet);
    procedure EndModal(Command: Word); virtual;
    function EventAvail: Boolean;
    function Execute: Word; virtual;
    function Exposed: Boolean;
    function Focus: Boolean;
    procedure GetBounds(var Bounds: TRect);
    procedure GetClipRect(var Clip: TRect);
    function GetColor(Color: Word): Word;
    procedure GetCommands(var Commands: TCommandSet);
    procedure GetData(var Rec); virtual;
    procedure GetEvent(var Event: TEvent); virtual;
    procedure GetExtent(var Extent: TRect);
    function GetHelpCtx: Word; virtual;
    function GetPalette: PPalette; virtual;
    function GetState(AState: Word): Boolean;
    procedure GrowTo(X, Y: Integer);
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Hide;
    procedure HideCursor;
    procedure KeyEvent(var Event: TEvent);
    procedure Locate(var Bounds: TRect);
    procedure MakeFirst;
    procedure MakeGlobal(Source: TPoint; var Dest: TPoint);
    procedure MakeLocal(Source: TPoint; var Dest: TPoint);
    function MouseEvent(var Event: TEvent; Mask: Word): Boolean;
    function MouseInView(Mouse: TPoint): Boolean;
    procedure MoveTo(X, Y: Integer);
    function NextView: PView;
    procedure NormalCursor;
    function Prev: PView;
    function PrevView: PView;
    procedure PutEvent(var Event: TEvent); virtual;
    procedure PutInFrontOf(Target: PView);
    procedure Select;
    procedure SetBounds(var Bounds: TRect);
    procedure SetCommands(Commands: TCommandSet);
    procedure SetCmdState(Commands: TCommandSet; Enable: Boolean);
    procedure SetCursor(X, Y: Integer);
    procedure SetData(var Rec); virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure Show;
    procedure ShowCursor;
    procedure SizeLimits(var Min, Max: TPoint); virtual;
    function TopView: PView;
    function Valid(Command: Word): Boolean; virtual;
    procedure WriteBuf(X, Y, W, H: Integer; var Buf);
    procedure WriteChar(X, Y: Integer; C: Char; Color: Byte;
      Count: Integer);
    procedure WriteLine(X, Y, W, H: Integer; var Buf);
    procedure WriteStr(X, Y: Integer; Str: String; Color: Byte);
  private
    procedure DrawCursor;
    procedure DrawHide(LastView: PView);
    procedure DrawShow(LastView: PView);
    procedure DrawUnderRect(var R: TRect; LastView: PView);
    procedure DrawUnderView(DoShadow: Boolean; LastView: PView);
    procedure ResetCursor; virtual;
    { --- Skelet --- }
    function  MapColor(color:byte):byte;
    procedure WriteView(x1,x2,y:Integer; var Buf);
    procedure WriteViewRec1(x1,x2:integer; p:PView; shadowCounter:integer);
    procedure WriteViewRec2(x1,x2:integer; p:PView; shadowCounter:integer);
    function  ExposedRec1(x1,x2:integer; p:PView):boolean;
    function  ExposedRec2(x1,x2:integer; p:PView):boolean;
  end;

{ TFrame types }

  TTitleStr = string[80];

{ TFrame object }

  { Palette layout }
  { 1 = Passive frame }
  { 2 = Passive title }
  { 3 = Active frame }
  { 4 = Active title }
  { 5 = Icons }

  PFrame = ^TFrame;
  TFrame = object(TView)
    constructor Init(var Bounds: TRect);
    procedure Draw; virtual;
    function GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
  private
    FrameMode: Word;
    procedure FrameLine(var FrameBuf; Y, N: Integer; Color: Byte);
  end;

{ ScrollBar characters }

  TScrollChars = array[0..4] of Char;

{ TScrollBar object }

  { Palette layout }
  { 1 = Page areas }
  { 2 = Arrows }
  { 3 = Indicator }

  PScrollBar = ^TScrollBar;
  TScrollBar = object(TView)
    Value: Integer;
    Min: Integer;
    Max: Integer;
    PgStep: Integer;
    ArStep: Integer;
    constructor Init(var Bounds: TRect);
    procedure Draw; virtual;
    function GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure ScrollDraw; virtual;
    function ScrollStep(Part: Integer): Integer; virtual;
    procedure SetParams(AValue, AMin, AMax, APgStep, AArStep: Integer);
    procedure SetRange(AMin, AMax: Integer);
    procedure SetStep(APgStep, AArStep: Integer);
    procedure SetValue(AValue: Integer);
  private
    Chars: TScrollChars;
    procedure DrawPos(Pos: Integer);
    function GetPos: Integer;
    function GetSize: Integer;
  end;

{ TScroller object }

  { Palette layout }
  { 1 = Normal text }
  { 2 = Selected text }

  PScroller = ^TScroller;
  TScroller = object(TView)
    HScrollBar: PScrollBar;
    VScrollBar: PScrollBar;
    Delta: TPoint;
    Limit: TPoint;
    constructor Init(var Bounds: TRect; AHScrollBar, AVScrollBar: PScrollBar);
    procedure ChangeBounds(var Bounds: TRect); virtual;
    function GetPalette: PPalette; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure ScrollDraw; virtual;
    procedure ScrollTo(X, Y: Integer);
    procedure SetLimit(X, Y: Integer);
    procedure SetState(AState: Word; Enable: Boolean); virtual;
  private
    DrawLock: Byte;
    DrawFlag: Boolean;
    procedure CheckDraw;
  end;

{ Video buffer }

  PVideoBuf = ^TVideoBuf;
  TVideoBuf = array[0..3999] of Word;

{ Selection modes }

  SelectMode = (NormalSelect, EnterSelect, LeaveSelect);

{ TGroup object }

  TGroup = object(TView)
    Last: PView;
    Current: PView;
    Phase: (phFocused, phPreProcess, phPostProcess);
    Buffer: PVideoBuf;
    EndState: Word;
    constructor Init(var Bounds: TRect);
    destructor Done; virtual;
    procedure Awaken; virtual;
    procedure ChangeBounds(var Bounds: TRect); virtual;
    function DataSize: Word; virtual;
    procedure Delete(P: PView);
    procedure Draw; virtual;
    procedure EndModal(Command: Word); virtual;
    procedure EventError(var Event: TEvent); virtual;
    function ExecView(P: PView): Word;
    function Execute: Word; virtual;
    function First: PView;
    function FirstThat(P: Pointer): PView;
    function FocusNext(Forwards: Boolean): Boolean;
    procedure ForEach(P: Pointer);
    procedure GetData(var Rec); virtual;
    function GetHelpCtx: Word; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Insert(P: PView);
    procedure InsertBefore(P, Target: PView);
    procedure Lock;
    procedure Redraw;
    procedure SelectNext(Forwards: Boolean);
    procedure SetData(var Rec); virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure Unlock;
    function Valid(Command: Word): Boolean; virtual;
  private
    Clip: TRect;
    LockFlag: Byte;
    function At(Index: Integer): PView;
    procedure DrawSubViews(P, Bottom: PView);
    function FirstMatch(AState: Word; AOptions: Word): PView;
    function FindNext(Forwards: Boolean): PView;
    procedure FreeBuffer;
    procedure GetBuffer;
    function IndexOf(P: PView): Integer;
    procedure InsertView(P, Target: PView);
    procedure RemoveView(P: PView);
    procedure ResetCurrent;
    procedure ResetCursor; virtual;
    procedure SetCurrent(P: PView; Mode: SelectMode);
  end;

{ TWindow object }

  { Palette layout }
  { 1 = Frame passive }
  { 2 = Frame active }
  { 3 = Frame icon }
  { 4 = ScrollBar page area }
  { 5 = ScrollBar controls }
  { 6 = Scroller normal text }
  { 7 = Scroller selected text }
  { 8 = Reserved }

  PWindow = ^TWindow;
  TWindow = object(TGroup)
    Flags: Byte;
    ZoomRect: TRect;
    Number: Integer;
    Palette: Integer;
    Frame: PFrame;
    Title: PString;
    constructor Init(var Bounds: TRect; ATitle: TTitleStr; ANumber: Integer);
    destructor Done; virtual;
    procedure Close; virtual;
    function GetPalette: PPalette; virtual;
    function GetTitle(MaxSize: Integer): TTitleStr; virtual;
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure InitFrame; virtual;
    procedure SetState(AState: Word; Enable: Boolean); virtual;
    procedure SizeLimits(var Min, Max: TPoint); virtual;
    function StandardScrollBar(AOptions: Word): PScrollBar;
    procedure Zoom; virtual;
  end;

{ Message dispatch function }

function Message(Receiver: PView; What, Command: Word;
  InfoPtr: Pointer): Pointer;

const

{ Event masks }

  PositionalEvents: Word = evMouse;
  FocusedEvents: Word = evKeyboard + evCommand;

{ Markers control }

  ShowMarkers: Boolean = False;

{ MapColor error return value }

  ErrorAttr: Byte = $CF;

{$IFNDEF OS_Linux}
  MinWinSize: TPoint = (X: 16; Y: 6);
  ShadowSize: TPoint = (X: 2; Y: 1);
{$ENDIF}
  ShadowAttr: Byte = $08;

{$IFDEF OS_Linux}
var
  MinWinSize: TPoint;
  ShadowSize: TPoint;
{$ENDIF}

const

{ Characters used for drawing selected and default items in  }
{ monochrome color sets                                      }

  SpecialChars: array[0..5] of Char = (#175, #174, #26, #27, ' ', ' ');

{ True if the command set has changed since being set to false }

  CommandSetChanged: Boolean = False;

implementation

uses Screen;

type
  PFixupList = ^TFixupList;
  TFixupList = array[1..4096] of Pointer;

const
  OwnerGroup: PGroup = nil;
  FixupList: PFixupList = nil;
  TheTopView: PView = nil;

const

{ Bit flags to determine how to draw the frame icons }

  fmCloseClicked = $0001;
  fmZoomClicked  = $0002;

{ Current command set. All but window commands are active by default }

  CurCommandSet: TCommandSet =
    [0..255] - [cmZoom, cmClose, cmResize, cmNext, cmPrev];

function TView.MapColor(color:byte):byte;
 var cur:PView; p:PPalette;
 begin
  if color=0 then MapColor:=errorAttr else begin
   cur:=@Self;
   repeat
    p:=cur^.GetPalette;
    if (p<>Nil) then if ord(p^[0])<>0 then begin
      if color>ord(p^[0]) then begin MapColor:=errorAttr; Exit end;
      color:=ord(p^[color]);
      if color=0 then begin MapColor:=errorAttr; Exit end;
    end;
    cur:=cur^.Owner;
   until (cur=Nil);
   MapColor:=color;
  end
 end;

function TView.GetColor(Color: Word): Word;
 var h:word;
 begin
  h:=Color shr 8; if h<>0 then h:=MapColor(h);
  GetColor:=(h shl 8) + MapColor(Color and $FF);
 end;

type
 PDrawBuffer = ^TDrawBuffer;
 TstatVar2 = record
  target:PView;
  offset,y:integer;
 end;

var
 staticVar1: PDrawBuffer;
 staticVar2: TstatVar2;

procedure doRefresh(p:PView);
 begin
 end;

procedure TView.writeViewRec1(x1,x2:integer; p:PView; shadowCounter:integer);
 var G:PGroup; BufPos,SrcPos,l,d,dx:word; TopScr:boolean;
 begin
  while true do begin
   p:=p^.Next;
   if (p=staticVar2.target) then begin
    G:=p^.Owner;
    if (G^.buffer<>Nil) then begin
{
/*
 * SS: now we should remove the mouse pointer from the screen.  This is
 * not necessary because we have a copy of the screen.
 *
 * if (p->owner->buffer == TScreen::screenBuffer) TScreen::drawMouse(0);
 */
}
     BufPos:=G^.size.x * staticVar2.y + x1;
     SrcPos:=x1 - staticVar2.offset; l:=x2-x1;
     TopScr:=G^.buffer=PVideoBuf(MyScreen.screenBuffer);
     if (shadowCounter=0) then begin
      {	/* SS: writes a row of data to the screen */ }
      if TopScr then MyScreen.writeRow(BufPos,staticVar1^[SrcPos],l);
      move(staticVar1^[SrcPos],G^.buffer^[BufPos],l shl 1);
     end else begin { paint with shadowAttr }
      while (l>0) do begin
       d:=staticVar1^[SrcPos] and ($FF or (shadowAttr shl 8));
       if TopScr then MyScreen.writeRow(BufPos,d,1);
       G^.buffer^[BufPos]:=d;
       inc(BufPos); inc(SrcPos); dec(l);
      end;
     end;
     { /* SS: draws mouse pointer */ }
     if TopScr then MyScreen.drawMouse(1);
    end; { G^.buffer<>Nil }
    if G^.lockFlag=0 then writeViewRec2(x1,x2,G,shadowCounter);
    exit;
   end; { p=staticVar2.target }

   if ((p^.state and sfVisible)<>0) and (staticVar2.y>=p^.Origin.Y) then begin

    if staticVar2.y<p^.Origin.Y+p^.size.Y then begin
     if x1<p^.origin.x then begin
      if x2<=p^.origin.x then continue;
      writeViewRec1(x1,p^.origin.x,p,shadowCounter);
      x1:=p^.origin.x;
     end;
     dx:=p^.origin.x+p^.size.x;
     if (x2<=dx) then exit; { return; }
     if (x1<dx) then x1:=dx;
     inc(dx,shadowSize.x);
     if ((p^.state and sfShadow)<>0) and (staticVar2.y>=p^.origin.y+shadowSize.y)
      then if (x1>dx) then continue else begin
        inc(shadowCounter);
        if (x2<=dx) then continue else begin
         writeViewRec1(x1,dx,p,shadowCounter);
         x1:=dx; dec(shadowCounter); continue;
        end;
      end else continue;
    end;

    if ((p^.state and sfShadow)<>0) and
       (staticVar2.y<p^.origin.y+p^.size.y+shadowSize.y) then begin
     dx:=p^.origin.x+shadowSize.x;
     if x1<dx then begin
      if x2<=dx then continue;
      writeViewRec1(x1,dx,p,shadowCounter);
      x1:=dx;
     end;
     inc(dx,p^.size.x);
     if x1>=dx then continue;
     inc(shadowCounter);
     if x2<=dx then continue else begin
      writeViewRec1(x1,dx,p,shadowCounter);
      x1:=dx; dec(shadowCounter);
     end
    end

   end;

  end; { while }
 end;

procedure TView.writeViewRec2(x1,x2:integer; p:PView; shadowCounter:integer);
 var savedStatics: TstatVar2; dx:integer; G:PGroup;
 begin
  G:=P^.Owner;
  if ((p^.State and sfVisible) <> 0) and (G<>Nil) then begin
   savedStatics:=staticVar2;
   inc(staticVar2.y,p^.Origin.Y);
   dx:=p^.Origin.X; inc(x1,dx); inc(x2,dx);
   inc(staticVar2.offset,dx);
   staticVar2.target:=p;
   if (staticVar2.y >= G^.clip.a.y) and (staticVar2.y < G^.clip.b.y) then begin
    if (x1<g^.clip.a.x) then x1 := g^.clip.a.x;
    if (x2>g^.clip.b.x) then x2 := g^.clip.b.x;
    if x1<x2 then writeViewRec1(x1,x2,G^.Last,shadowCounter);
   end;
   staticVar2 := savedStatics;
  end;
 end;

procedure TView.WriteView(x1,x2,y:integer; var Buf);
 begin if (y>=0) and (y<Size.Y) then begin
  if x1<0 then x1:=0;
  if x2>Size.X then x2:=Size.X;
  if x1<x2 then begin
   staticVar2.offset:=x1;
   staticVar2.y:=y;
   staticVar1:=@Buf;
   writeViewRec2( x1, x2, @Self, 0 );
   doRefresh(@Self);
  end;
 end end;

{ TView }

constructor TView.Init(var Bounds: TRect);
 begin
  TObject.Init;
  Owner := nil;
  State := sfVisible;
  SetBounds(Bounds);
  DragMode := dmLimitLoY;
  HelpCtx := hcNoContext;
  EventMask := evMouseDown + evKeyDown + evCommand;
 end;

destructor TView.Done;
 begin Hide; if Owner <> nil then Owner^.Delete(@Self) end;

procedure TView.Awaken;
 begin end;

procedure TView.BlockCursor;
 begin SetState(sfCursorIns, True) end;

procedure TView.CalcBounds(var Bounds: TRect; Delta: TPoint);
 var S, D: Integer; Min, Max: TPoint;
 procedure Grow(var I: Integer);
  begin
   if GrowMode and gfGrowRel = 0
    then Inc(I, D)
    else I := (I * S + (S - D) shr 1) div (S - D);
  end;
 function Range(Val, Min, Max: Integer): Integer;
  begin
   if Val < Min
    then Range := Min
    else if Val > Max then Range := Max else Range := Val;
  end;
 begin
  GetBounds(Bounds);
  S := Owner^.Size.X;
  D := Delta.X;
  if GrowMode and gfGrowLoX <> 0 then Grow(Bounds.A.X);
  if GrowMode and gfGrowHiX <> 0 then Grow(Bounds.B.X);
  if Bounds.B.X - Bounds.A.X > MaxViewWidth
   then Bounds.B.X := Bounds.A.X + MaxViewWidth;
  S := Owner^.Size.Y;
  D := Delta.Y;
  if GrowMode and gfGrowLoY <> 0 then Grow(Bounds.A.Y);
  if GrowMode and gfGrowHiY <> 0 then Grow(Bounds.B.Y);
  SizeLimits(Min, Max);
  Bounds.B.X := Bounds.A.X + Range(Bounds.B.X - Bounds.A.X, Min.X, Max.X);
  Bounds.B.Y := Bounds.A.Y + Range(Bounds.B.Y - Bounds.A.Y, Min.Y, Max.Y);
 end;

procedure TView.ChangeBounds(var Bounds: TRect);
 begin SetBounds(Bounds); DrawView end;

procedure TView.ClearEvent(var Event: TEvent);
 begin
  Event.What := evNothing;
  Event.InfoPtr := @Self;
 end;

function TView.CommandEnabled(Command: Word): Boolean;
 begin
  CommandEnabled := (Command > 255) or (Command in CurCommandSet);
 end;

function TView.DataSize: Word; begin DataSize := 0 end;

procedure TView.DisableCommands(Commands: TCommandSet);
 begin
  CommandSetChanged := CommandSetChanged or (CurCommandSet * Commands <> []);
  CurCommandSet := CurCommandSet - Commands;
 end;

procedure TView.DragView(Event: TEvent; Mode: Byte;
  var Limits: TRect; MinSize, MaxSize: TPoint);
 var P, S: TPoint; SaveBounds: TRect;
 function Min(I, J: Integer): Integer;
  begin if I < J then Min := I else Min := J end;
 function Max(I, J: Integer): Integer;
  begin if I > J then Max := I else Max := J end;
 procedure MoveGrow(P, S: TPoint);
  var R: TRect;
  begin
   S.X := Min(Max(S.X, MinSize.X), MaxSize.X);
   S.Y := Min(Max(S.Y, MinSize.Y), MaxSize.Y);
   P.X := Min(Max(P.X, Limits.A.X - S.X + 1), Limits.B.X - 1);
   P.Y := Min(Max(P.Y, Limits.A.Y - S.Y + 1), Limits.B.Y - 1);
   if Mode and dmLimitLoX <> 0 then P.X := Max(P.X, Limits.A.X);
   if Mode and dmLimitLoY <> 0 then P.Y := Max(P.Y, Limits.A.Y);
   if Mode and dmLimitHiX <> 0 then P.X := Min(P.X, Limits.B.X - S.X);
   if Mode and dmLimitHiY <> 0 then P.Y := Min(P.Y, Limits.B.Y - S.Y);
   R.Assign(P.X, P.Y, P.X + S.X, P.Y + S.Y);
   Locate(R);
  end;
 procedure Change(DX, DY: Integer);
  begin
   if (Mode and dmDragMove <> 0) and (GetShiftState and $03 = 0)
    then begin Inc(P.X, DX); Inc(P.Y, DY) end
    else if (Mode and dmDragGrow <> 0) and (GetShiftState and $03 <> 0)
     then begin Inc(S.X, DX); Inc(S.Y, DY) end;
  end;
 procedure Update(X, Y: Integer);
  begin
   if Mode and dmDragMove<>0 then begin P.X := X; P.Y := Y end;
  end;
 begin
  SetState(sfDragging, True);
  if Event.What = evMouseDown then begin
   if Mode and dmDragMove <> 0 then begin
    P.X := Origin.X - Event.Where.X;
    P.Y := Origin.Y - Event.Where.Y;
    repeat
     TPointInc(Event.Where, P);
     MoveGrow(Event.Where, Size);
    until not MouseEvent(Event, evMouseMove);
   end else begin
    P.X := Size.X - Event.Where.X;
    P.Y := Size.Y - Event.Where.Y;
    repeat
     TPointInc(Event.Where, P);
     MoveGrow(Origin, Event.Where);
    until not MouseEvent(Event, evMouseMove);
   end;
  end else begin
   GetBounds(SaveBounds);
   repeat
    P := Origin; S := Size;
    KeyEvent(Event);
    case Event.KeyCode and $FF00 of
     kbLeft: Change(-1, 0);
     kbRight: Change(1, 0);
     kbUp: Change(0, -1);
     kbDown: Change(0, 1);
     kbCtrlLeft: Change(-8, 0);
     kbCtrlRight: Change(8, 0);
     kbHome: Update(Limits.A.X, P.Y);
     kbEnd: Update(Limits.B.X - S.X, P.Y);
     kbPgUp: Update(P.X, Limits.A.Y);
     kbPgDn: Update(P.X, Limits.B.Y - S.Y);
    end;
    MoveGrow(P, S);
   until (Event.KeyCode = kbEnter) or (Event.KeyCode = kbEsc);
   if Event.KeyCode = kbEsc then Locate(SaveBounds);
  end;
  SetState(sfDragging, False);
 end;

procedure TView.Draw;
 var B:TDrawBuffer;
 begin
  MoveChar(B, ' ', GetColor(1), Size.X);
  WriteLine(0, 0, Size.X, Size.Y, B);
 end;

procedure TView.DrawCursor;
 begin if State and sfFocused <> 0 then ResetCursor end;

procedure TView.DrawHide(LastView: PView);
 begin
  DrawCursor;
  DrawUnderView(State and sfShadow <> 0, LastView);
 end;

procedure TView.DrawShow(LastView: PView);
 begin
  DrawView;
  if State and sfShadow <> 0 then DrawUnderView(True, LastView);
 end;

procedure TView.DrawUnderRect(var R: TRect; LastView: PView);
 begin
  Owner^.Clip.Intersect(R);
  Owner^.DrawSubViews(NextView, LastView);
  Owner^.GetExtent(Owner^.Clip);
 end;

procedure TView.DrawUnderView(DoShadow: Boolean; LastView: PView);
 var R: TRect;
 begin
  GetBounds(R);
  if DoShadow then TPointInc(R.B,ShadowSize);
  DrawUnderRect(R, LastView);
 end;

procedure TView.DrawView;
 begin if Exposed then begin Draw; DrawCursor end end;

procedure TView.EnableCommands(Commands: TCommandSet);
 begin
  CommandSetChanged :=
   CommandSetChanged or (CurCommandSet * Commands <> Commands);
  CurCommandSet := CurCommandSet + Commands;
 end;

procedure TView.EndModal(Command: Word);
 var P: PView;
 begin
  P := TopView;
  if TopView <> nil then TopView^.EndModal(Command);
 end;

function TView.EventAvail: Boolean;
 var Event: TEvent;
 begin
  GetEvent(Event);
  if Event.What <> evNothing then PutEvent(Event);
  EventAvail := Event.What <> evNothing;
 end;

procedure TView.GetBounds(var Bounds: TRect);
 begin with Bounds do begin
  A:=Origin; B:=A;
  TPointInc(B,Size);
 end end;

function TView.Execute: Word;
 begin Execute := cmCancel end;

function TView.ExposedRec1(x1,x2:integer; p:PView):boolean;
 var G:PGroup; dy,dx:integer;
 begin while true do begin
  p:=p^.Next; G:=p^.Owner;
  if p=staticVar2.target
   then begin exposedRec1:=exposedRec2(x1,x2,G); Exit end;
  dy:=p^.origin.y; dx:=p^.origin.x;
  if ((p^.state and sfVisible)<>0) and (staticVar2.y>=dy) then begin
   if staticVar2.y<dy+p^.size.y then begin
    if x1<dx then begin
     if x2<=dx then continue;
     if x2>dx+p^.size.x then begin
      if exposedRec1(x1,dx,p) then begin exposedRec1:=True; Exit end;
      x1:=dx+p^.size.x;
     end else x2:=dx;
    end else begin
     if x1<dx+p^.size.x then x1:=dx+p^.size.x;
     if x1>=x2 then begin exposedRec1:=False; Exit end;
    end;
   end;
  end;
 end end;

function TView.ExposedRec2(x1,x2:integer; p:PView):boolean;
 var G:PGroup; savedStat:TStatVar2;
 begin
  if (p^.state and sfVisible)=0 then ExposedRec2:=false else begin
   G:=p^.Owner;
   if (G=Nil) or (G^.Buffer<>Nil) then ExposedRec2:=true else begin
    savedStat:=staticVar2;
    inc(staticVar2.y,p^.origin.y);
    inc(x1,p^.origin.x);
    inc(x2,p^.origin.x);
    staticVar2.target:=p;
    if (staticVar2.y<G^.clip.a.y) or
       (staticVar2.y>=G^.clip.b.y)
     then ExposedRec2:=false
     else begin
      if (x1<G^.clip.a.x) then x1:=G^.clip.a.x;
      if (x2>G^.clip.b.x) then x2:=G^.clip.b.x;
      if (x1>=x2)
       then ExposedRec2:=false
       else ExposedRec2:=exposedRec1(x1,x2,G^.Last);
     end;
    staticVar2 := savedStat;
   end
  end
 end;

function TView.Exposed: Boolean;
 var OK:boolean; y:word;
 begin
  if ((State and sfExposed)<>0) and (Size.X>0) and (Size.Y>0) then begin
   OK:=false; y:=0;
   while (y<Size.Y) and (not OK) do begin
    staticVar2.y:=y;
    OK:=ExposedRec2(0,Size.X,@Self);
    inc(y);
   end;
   Exposed:=OK;
  end else Exposed:=False
 end;

function TView.Focus: Boolean;
 var Result: Boolean;
 begin
  Result := True;
  if State and (sfSelected + sfModal) = 0 then begin
   if Owner <> nil then begin
    Result := Owner^.Focus;
    if Result
     then if ((Owner^.Current = nil) or
       (Owner^.Current^.Options and ofValidate = 0) or
       (Owner^.Current^.Valid(cmReleasedFocus)))
      then Select else Result := False;
   end;
  end;
  Focus := Result;
 end;

procedure TView.GetClipRect(var Clip: TRect);
 begin
  GetBounds(Clip);
  if Owner <> nil then Clip.Intersect(Owner^.Clip);
  Clip.Move(-Origin.X, -Origin.Y);
 end;

procedure TView.GetCommands(var Commands: TCommandSet);
 begin Commands := CurCommandSet end;

procedure TView.GetData(var Rec); begin end;

procedure TView.GetEvent(var Event: TEvent);
 begin if Owner <> nil then Owner^.GetEvent(Event) end;

procedure TView.GetExtent(var Extent: TRect);
 begin with Extent do begin A.X:=0; A.Y:=0; B:=Size end end;

function TView.GetHelpCtx: Word;
 begin
  if State and sfDragging <> 0
   then GetHelpCtx := hcDragging
   else GetHelpCtx := HelpCtx;
 end;

function TView.GetPalette: PPalette;
 begin GetPalette := nil end;

function TView.GetState(AState: Word): Boolean;
 begin GetState := State and AState = AState end;

procedure TView.GrowTo(X, Y: Integer);
 var R: TRect;
 begin
  R.Assign(Origin.X, Origin.Y, Origin.X + X, Origin.Y + Y);
  Locate(R);
 end;

procedure TView.HandleEvent(var Event: TEvent);
 begin
  if Event.What = evMouseDown
   then if (State and (sfSelected + sfDisabled) = 0) and
     (Options and ofSelectable <> 0)
    then if not Focus or (Options and ofFirstClick = 0)
     then ClearEvent(Event);
 end;

procedure TView.Hide;
 begin
  if State and sfVisible <> 0 then SetState(sfVisible, False);
 end;

procedure TView.HideCursor;
 begin SetState(sfCursorVis, False) end;

procedure TView.KeyEvent(var Event: TEvent);
 begin repeat GetEvent(Event) until Event.What = evKeyDown end;

procedure TView.Locate(var Bounds: TRect);
 var R: TRect; Min, Max: TPoint;
 function Range(Val, Min, Max: Integer): Integer;
  begin
   if Val < Min
    then Range := Min
    else if Val > Max then Range := Max else Range := Val;
  end;
 begin
  SizeLimits(Min, Max);
  Bounds.B.X := Bounds.A.X + Range(Bounds.B.X - Bounds.A.X, Min.X, Max.X);
  Bounds.B.Y := Bounds.A.Y + Range(Bounds.B.Y - Bounds.A.Y, Min.Y, Max.Y);
  GetBounds(R);
  if not Bounds.Equals(R) then begin
   ChangeBounds(Bounds);
   if (Owner <> nil) and (State and sfVisible <> 0) then begin
    if State and sfShadow <> 0 then begin
     R.Union(Bounds);
     TPointInc(R.B, ShadowSize);
    end;
    DrawUnderRect(R, nil);
   end;
  end;
 end;

procedure TView.MakeFirst;
 begin PutInFrontOf(Owner^.First) end;

procedure TView.MakeGlobal(Source: TPoint; var Dest: TPoint);
 var cur:PView;
 begin
  cur:=@Self; Dest:=Source;
  TPointInc(Dest,Origin);
  while cur^.Owner<>Nil do begin
   cur:=cur^.Owner;
   TPointInc(Dest,cur^.Origin);
  end
 end;

procedure TView.MakeLocal(Source: TPoint; var Dest: TPoint);
 var cur:PView;
 begin
  cur:=@Self; Dest:=Source;
  TPointDec(Dest,Origin);
  while cur^.Owner<>Nil do begin
   cur:=cur^.Owner;
   TPointDec(Dest,cur^.Origin);
  end
 end;

function TView.MouseEvent(var Event: TEvent; Mask: Word): Boolean;
 begin
  repeat GetEvent(Event) until Event.What and (Mask or evMouseUp) <> 0;
  MouseEvent := Event.What <> evMouseUp;
 end;

function TView.MouseInView(Mouse: TPoint): Boolean;
 var Extent: TRect;
 begin
  MakeLocal(Mouse, Mouse);
  GetExtent(Extent);
  MouseInView := Extent.Contains(Mouse);
 end;

procedure TView.MoveTo(X, Y: Integer);
 var R: TRect;
 begin
  R.Assign(X, Y, X + Size.X, Y + Size.Y);
  Locate(R);
 end;

function TView.NextView: PView;
 begin
  if @Self = Owner^.Last then NextView := nil else NextView := Next;
 end;

procedure TView.NormalCursor;
 begin SetState(sfCursorIns, False) end;

function TView.Prev: PView;
 var res:PView;
 begin
  res:=@Self;
  while res^.Next<>@Self do res:=res^.Next;
  Prev:=res;
 end;

function TView.PrevView: PView;
 begin
  if @Self = Owner^.First then PrevView := nil else PrevView := Prev;
 end;

procedure TView.PutEvent(var Event: TEvent);
 begin if Owner <> nil then Owner^.PutEvent(Event) end;

procedure TView.PutInFrontOf(Target: PView);
 var P, LastView: PView;
 procedure MoveView;
  begin
   Owner^.RemoveView(@Self);
   Owner^.InsertView(@Self, Target);
 end;
 begin
  if (Owner <> nil) and (Target <> @Self) and (Target <> NextView) and
    ((Target = nil) or (Target^.Owner = Owner))
   then if State and sfVisible = 0 then MoveView else begin
    LastView := NextView;
    if LastView <> nil then begin
     P := Target;
     while (P <> nil) and (P <> LastView) do P := P^.NextView;
     if P = nil then LastView := Target;
    end;
    State := State and not sfVisible;
    if LastView = Target then DrawHide(LastView);
    MoveView;
    State := State or sfVisible;
    if LastView <> Target then DrawShow(LastView);
    if Options and ofSelectable <> 0 then begin
     Owner^.ResetCurrent;
     Owner^.ResetCursor;
    end;
   end;
 end;

procedure TView.ResetCursor;
 const sfV_CV_F:word = sfVisible + sfCursorVis + sfFocused;
 var p,p2:PView; G:PGroup; cur:TPoint;
 function Check0:boolean;
  var res:byte;
  begin
   res:=0;
   while res=0 do begin
    p:=p^.next;
    if p=p2
     then begin p:=P^.owner; res:=1 end
     else if ((p^.state and sfVisible)<>0) and
             (cur.x>=p^.origin.x) and
             (cur.x<p^.size.x+p^.origin.x) and
             (cur.y>=p^.origin.y) and
             (cur.y<p^.size.y+p^.origin.y) then res:=2;
   end;
   Check0:=res=2;
  end;
 begin
  if (state and sfV_CV_F) = sfV_CV_F then begin
   p:=@Self;
   cur:=cursor;
   while true do begin
    if (cur.x<0) or (cur.x>=p^.size.x) or
       (cur.y<0) or (cur.y>=p^.size.y) then break;
    TPointInc(cur,p^.origin);
    p2:=p; G:=p^.owner;
    if { G^.buffer=PVideoBuf(MyScreen.screenBuffer) }
       G=Nil then begin { top view }
{        /*
         * SS: we should change the cursor size according to the sfCursorIns
         * flag in the state variable:
         *
         * if (state and sfCursorIns)<>0
         *  then setBigCursor
         *  else setSmallCursor;
         *
         * Is there a way to do it under linux ?
         */
}
     MyScreen.moveCursor(cur.x,cur.y);
     MyScreen.drawCursor(true,(state and sfCursorIns)<>0);
     exit;
    end;
    if (G^.state and sfVisible)=0 then break;
    p:=G^.Last;
    if Check0 then break;
   end; { while }
  end; { if }
  MyScreen.drawCursor(false,false);
 end;

procedure TView.Select;
 begin
  if Options and ofSelectable <> 0
   then if Options and ofTopSelect <> 0
    then MakeFirst
    else if Owner <> nil then Owner^.SetCurrent(@Self, NormalSelect);
 end;

procedure TView.SetBounds(var Bounds: TRect);
 begin
  origin := bounds.a; size := bounds.b;
  TPointDec(size,origin);
 end;

procedure TView.SetCmdState(Commands: TCommandSet; Enable: Boolean);
 begin
  if Enable
   then EnableCommands(Commands)
   else DisableCommands(Commands);
 end;

procedure TView.SetCommands(Commands: TCommandSet);
 begin
  CommandSetChanged := CommandSetChanged or (CurCommandSet <> Commands);
  CurCommandSet := Commands;
 end;

procedure TView.SetCursor(X, Y: Integer);
 begin
  Cursor.X := X;
  Cursor.Y := Y;
  DrawCursor;
 end;

procedure TView.SetData(var Rec); begin end;

procedure TView.SetState(AState: Word; Enable: Boolean);
 var Command: Word;
 begin
  if Enable
   then State := State or AState
   else State := State and not AState;
  if Owner <> nil then case AState of
   sfVisible: begin
     if Owner^.State and sfExposed <> 0 then SetState(sfExposed, Enable);
     if Enable then DrawShow(nil) else DrawHide(nil);
     if Options and ofSelectable <> 0 then Owner^.ResetCurrent;
    end;
   sfCursorVis, sfCursorIns: DrawCursor;
   sfShadow: DrawUnderView(True, nil);
   sfFocused: begin
     ResetCursor;
     if Enable
      then Command := cmReceivedFocus
      else Command := cmReleasedFocus;
     Message(Owner, evBroadcast, Command, @Self);
    end;
  end;
 end;

procedure TView.Show;
 begin
  if State and sfVisible = 0 then SetState(sfVisible, True);
 end;

procedure TView.ShowCursor;
 begin SetState(sfCursorVis, True) end;

procedure TView.SizeLimits(var Min, Max: TPoint);
 begin
  Min.X:=0; Min.Y:=0;
  if Owner <> nil
   then Max := Owner^.Size
   else with Max do begin X:=$7FFF; Y:=$7FFF end;
 end;

function TView.TopView: PView;
 var P: PView;
 begin
  if TheTopView = nil then begin
   P := @Self;
   while (P <> nil) and (P^.State and sfModal = 0) do P := P^.Owner;
   TopView := P;
  end else TopView := TheTopView;
 end;

function TView.Valid(Command: Word): Boolean;
 begin Valid := True end;

procedure TView.WriteBuf(X, Y, W, H: Integer; var Buf);
 var i:integer;
 begin
  if h>0 then for i:= 0 to h-1 do
   writeView(X,X+W,Y+i,TVideoBuf(Buf)[W*i])
 end;

procedure TView.WriteChar(X,Y:Integer; C:Char; Color:Byte; Count:Integer);
 var B:TDrawBuffer; myChar,i:word;
 begin
  myChar:=MapColor(Color);
  myChar:=(myChar shl 8) + ord(C);
  if Count>0 then begin
   if Count>maxViewWidth then Count:=maxViewWidth;
   for i:=0 to Count-1 do B[i]:=myChar;
   writeView(X,X+Count,Y,B);
  end;
 end;

procedure TView.WriteLine(X, Y, W, H: Integer; var Buf);
 var i:integer;
 begin if h>0 then for i:=0 to h-1 do writeView(x,x+w,y+i,buf) end;

procedure TView.WriteStr(X, Y: Integer; Str: String; Color: Byte);
 var l,i:word; B:TDrawBuffer; myColor:word;
 begin
  l:=length(Str);
  if l>0 then begin
   if l>maxViewWidth then l:=maxViewWidth;
   MyColor:=MapColor(Color); MyColor:=MyColor shl 8;
   for i:=0 to l-1 do B[i]:=MyColor+ord(Str[i+1]);
   writeView(x,x+l,y,b);
  end
 end;

{ TFrame }

constructor TFrame.Init(var Bounds: TRect);
 begin
  TView.Init(Bounds);
  GrowMode := gfGrowHiX + gfGrowHiY;
  EventMask := EventMask or evBroadcast;
 end;

{
Erklrung der Mask:

       Bit 0
         |
  Bit 3 - - Bit 1
         |
       Bit 2

Wenn z.B. sichergestellt werden soll, dass eine linke obere Ecke im
Muster vorhanden ist, nimmt man :
  mask |= 0x06 .
}

procedure TFrame.FrameLine(var FrameBuf; Y, N: Integer; Color: Byte);
 const
  InitFrame: array[0..17] of Byte =
    ($06, $0A, $0C, $05, $00, $05, $03, $0A, $09,
     $16, $1A, $1C, $15, $00, $15, $13, $1A, $19);
  FrameChars: array[0..31] of Char =
{$IFDEF Arabian}
    '           ';
{$ELSE}
    '          ϻ ';
{$ENDIF}
 var
  FrameMask:array[0..MaxViewWidth-1] of Byte;
  i:integer; p:PView; mask1,mask2:byte; wCol:word;
  xMin,xMax,yMin,yMax:integer;
 begin
  frameMask[0]:=initFrame[n];
  for i:=1 to size.X-2 do frameMask[i]:=initFrame[n+1];
  frameMask[size.x-1]:=initFrame[n+2];
  p:=Owner^.Last;
  while true do begin
   p:=p^.Next;
   if p=@Self then break;
   if ((p^.options and ofFramed)<>0) and
      ((p^.state and sfVisible)<>0) then begin
    yMin:=p^.origin.y; yMax:=yMin+p^.size.y;
    if y+1<yMin then continue
     else if y+1=yMin then begin mask1:=$0A; mask2:=$06 end
      else if y=yMax  then begin mask1:=$0A; mask2:=$03 end
       else if y<yMax then begin mask1:=$00; mask2:=$05 end
        else continue;
    xMin:=p^.origin.x; xMax:=xMin+p^.size.x;
    if xMin<1 then xMin:=1;
    if xMax>size.x-1 then xMax:=size.x-1;
    if xMax>xMin then begin
     frameMask[xMin-1]:=frameMask[xMin-1] or mask2;
     frameMask[xMax]:=frameMask[xMax] or (mask2 xor mask1);
     if mask1<>0 then for i:=xMin to xMax-1 do
      frameMask[i]:=frameMask[i] or mask1;
    end;
   end; { if Framed and Visible }
  end; { while }
  wCol:=Color; wCol:=wCol shl 8;
  for i:=0 to size.X-1 do
   TDrawBuffer(frameBuf)[i]:=wCol+ord(frameChars[frameMask[i]]);
 end;

procedure TFrame.Draw;
 var
  CFrame, CTitle: Word;
  F, I, L, Width: Integer;
  B: TDrawBuffer;
  Title: TTitleStr;
  Min, Max: TPoint;
 begin
  if State and sfDragging <> 0
   then begin CFrame := $0505; CTitle := $0005; F := 0 end
   else if State and sfActive = 0
    then begin CFrame := $0101; CTitle := $0002; F := 0 end
    else begin CFrame := $0503; CTitle := $0004; F := 9 end;
  CFrame := GetColor(CFrame);
  CTitle := GetColor(CTitle);
  Width := Size.X;
  L := Width - 10;
  if PWindow(Owner)^.Flags and (wfClose+wfZoom) <> 0 then Dec(L,6);
  FrameLine(B, 0, F, Byte(CFrame));
  if (PWindow(Owner)^.Number <> wnNoNumber) and
     (PWindow(Owner)^.Number < 10)
   then begin
    Dec(L,4);
    if PWindow(Owner)^.Flags and wfZoom <> 0
     then I := 7 else I := 3;
    WordRec(B[Width - I]).Lo := PWindow(Owner)^.Number + $30;
   end;
  if Owner <> nil
   then Title := PWindow(Owner)^.GetTitle(L)
   else Title := '';
  if Title <> '' then begin
   L := Length(Title);
   if L > Width - 10 then L := Width - 10;
   if L < 0 then L := 0;
   I := (Width - L) shr 1;
   MoveChar(B[I - 1], ' ', CTitle, 1);
   MoveBuf(B[I], Title[1], CTitle, L);
   MoveChar(B[I + L], ' ', CTitle, 1);
  end;
  if State and sfActive <> 0 then begin
   if PWindow(Owner)^.Flags and wfClose <> 0
    then if FrameMode and fmCloseClicked = 0
     then MoveCStr(B[2], '[~~]', CFrame)
     else MoveCStr(B[2], '[~'#15'~]', CFrame);
   if PWindow(Owner)^.Flags and wfZoom <> 0 then begin
    MoveCStr(B[Width - 5], '[~'#24'~]', CFrame);
    Owner^.SizeLimits(Min, Max);
    if FrameMode and fmZoomClicked <> 0
     then WordRec(B[Width - 4]).Lo := 15
     else if (Owner^.Size.X = Max.X) and (Owner^.Size.Y = Max.Y)
      then WordRec(B[Width - 4]).Lo := 18;
   end;
  end;
  WriteLine(0, 0, Size.X, 1, B);
  for I := 1 to Size.Y - 2 do begin
    FrameLine(B, I, F + 3, Byte(CFrame));
    WriteLine(0, I, Size.X, 1, B);
  end;
  FrameLine(B, Size.Y - 1, F + 6, Byte(CFrame));
  if State and sfActive <> 0
   then if PWindow(Owner)^.Flags and wfGrow <> 0
    then MoveCStr(B[Width - 2], '~~', CFrame);
  WriteLine(0, Size.Y - 1, Size.X, 1, B);
 end;

function TFrame.GetPalette: PPalette;
 const P: String[Length(CFrame)] = CFrame;
 begin GetPalette := @P end;

procedure TFrame.HandleEvent(var Event: TEvent);
 var Mouse: TPoint;
 procedure DragWindow(Mode: Byte);
  var Limits: TRect; Min, Max: TPoint;
  begin
   Owner^.Owner^.GetExtent(Limits);
   Owner^.SizeLimits(Min, Max);
   Owner^.DragView(Event, Owner^.DragMode or Mode, Limits, Min, Max);
   ClearEvent(Event);
  end;
 begin
  TView.HandleEvent(Event);
  if Event.What = evMouseDown then begin
   MakeLocal(Event.Where, Mouse);
   if Mouse.Y = 0 then begin
    if (PWindow(Owner)^.Flags and wfClose <> 0) and
      (State and sfActive <> 0) and (Mouse.X >= 2) and (Mouse.X <= 4)
     then begin
      repeat
       MakeLocal(Event.Where, Mouse);
       if (Mouse.X >= 2) and (Mouse.X <= 4) and (Mouse.Y = 0)
        then FrameMode := fmCloseClicked
        else FrameMode := 0;
       DrawView;
      until not MouseEvent(Event, evMouseMove + evMouseAuto);
      FrameMode := 0;
      if (Mouse.X >= 2) and (Mouse.X <= 4) and (Mouse.Y = 0) then begin
       Event.What := evCommand;
       Event.Command := cmClose;
       Event.InfoPtr := Owner;
       PutEvent(Event);
      end;
      ClearEvent(Event);
      DrawView;
     end
     else if (PWindow(Owner)^.Flags and wfZoom <> 0) and
       (State and sfActive <> 0) and (Event.Double or
       (Mouse.X >= Size.X - 5) and
       (Mouse.X <= Size.X - 3))
      then begin
       if not Event.Double then repeat
        MakeLocal(Event.Where, Mouse);
        if (Mouse.X >= Size.X - 5) and (Mouse.X <= Size.X - 3) and
           (Mouse.Y = 0)
         then FrameMode := fmZoomClicked
         else FrameMode := 0;
        DrawView;
       until not MouseEvent(Event, evMouseMove + evMouseAuto);
       FrameMode := 0;
       if ((Mouse.X >= Size.X - 5) and (Mouse.X <= Size.X - 3) and
         (Mouse.Y = 0)) or Event.Double
        then begin
         Event.What := evCommand;
         Event.Command := cmZoom;
         Event.InfoPtr := Owner;
         PutEvent(Event);
        end;
       ClearEvent(Event);
       DrawView;
      end
      else if PWindow(Owner)^.Flags and wfMove <> 0
       then DragWindow(dmDragMove);
   end else if (State and sfActive <> 0) and
     (Mouse.X >= Size.X - 2) and
     (Mouse.Y >= Size.Y - 1)
    then if PWindow(Owner)^.Flags and wfGrow <> 0
     then DragWindow(dmDragGrow);
  end;
 end;

procedure TFrame.SetState(AState: Word; Enable: Boolean);
 begin
  TView.SetState(AState, Enable);
  if AState and (sfActive + sfDragging) <> 0 then DrawView;
 end;

{ TScrollBar }

constructor TScrollBar.Init(var Bounds: TRect);
 const
  VChars: TScrollChars = (#30, #31, #177, #254, #178);
  HChars: TScrollChars = (#17, #16, #177, #254, #178);
 begin
  TView.Init(Bounds);
  Value := 0; Min := 0; Max := 0;
  PgStep := 1; ArStep := 1;
  if Size.X = 1 then begin
   GrowMode := gfGrowLoX + gfGrowHiX + gfGrowHiY;
   Chars := VChars;
  end else begin
   GrowMode := gfGrowLoY + gfGrowHiX + gfGrowHiY;
   Chars := HChars;
  end;
 end;

procedure TScrollBar.Draw;
 begin DrawPos(GetPos) end;

procedure TScrollBar.DrawPos(Pos: Integer);
 var S: Integer; B: TDrawBuffer;
 begin
  S := GetSize - 1;
  MoveChar(B[0], Chars[0], GetColor(2), 1);
  if Max = Min
   then MoveChar(B[1], Chars[4], GetColor(1), S - 1)
   else begin
    MoveChar(B[1], Chars[2], GetColor(1), S - 1);
    MoveChar(B[Pos], Chars[3], GetColor(3), 1);
   end;
  MoveChar(B[S], Chars[1], GetColor(2), 1);
  WriteBuf(0, 0, Size.X, Size.Y, B);
 end;

function TScrollBar.GetPalette: PPalette;
 const P: String[Length(CScrollBar)] = CScrollBar;
 begin GetPalette := @P end;

function TScrollBar.GetPos: Integer;
 var R: Integer;
 begin
  R := Max - Min;
  if R = 0
   then GetPos := 1
   else GetPos := LongDiv(LongMul(Value - Min, GetSize - 3) + R shr 1, R) + 1;
 end;

function TScrollBar.GetSize: Integer;
 var S: Integer;
 begin
  if Size.X = 1 then S := Size.Y else S := Size.X;
  if S < 3 then GetSize := 3 else GetSize := S;
 end;

procedure TScrollBar.HandleEvent(var Event: TEvent);
 var
  Tracking: Boolean;
  I, P, S, ClickPart: Integer;
  Mouse: TPoint;
  Extent: TRect;
 function GetPartCode: Integer;
  var Mark, Part: Integer;
  begin
   Part := -1;
   if Extent.Contains(Mouse) then begin
    if Size.X = 1 then Mark := Mouse.Y else Mark := Mouse.X;
    if Mark = P then Part := sbIndicator else begin
     if Mark < 1 then Part := sbLeftArrow else
      if Mark < P then Part := sbPageLeft else
       if Mark < S then Part := sbPageRight else Part := sbRightArrow;
     if Size.X = 1 then Inc(Part, 4);
    end;
   end;
   GetPartCode := Part;
  end;
 procedure Clicked;
  begin
   Message(Owner, evBroadcast, cmScrollBarClicked, @Self);
  end;
 begin
  TView.HandleEvent(Event);
  case Event.What of
   evMouseDown: begin
     Clicked;
     MakeLocal(Event.Where, Mouse);
     GetExtent(Extent);
     Extent.Grow(1, 1);
     P := GetPos;
     S := GetSize - 1;
     ClickPart := GetPartCode;
     if ClickPart <> sbIndicator then begin
      repeat
       MakeLocal(Event.Where, Mouse);
       if GetPartCode = ClickPart
        then SetValue(Value + ScrollStep(ClickPart));
      until not MouseEvent(Event, evMouseAuto);
     end else begin
      repeat
       MakeLocal(Event.Where, Mouse);
       Tracking := Extent.Contains(Mouse);
       if Tracking then begin
        if Size.X = 1 then I := Mouse.Y else I := Mouse.X;
        if I <= 0 then I := 1;
        if I >= S then I := S - 1;
       end else I := GetPos;
       if I <> P then begin DrawPos(I); P := I end;
      until not MouseEvent(Event, evMouseMove);
      if Tracking and (S > 2) then begin
       Dec(S, 2);
       SetValue(LongDiv(LongMul(P - 1, Max - Min) + S shr 1, S) + Min);
      end;
     end;
     ClearEvent(Event);
    end;
   evKeyDown: if State and sfVisible <> 0 then begin
     ClickPart := sbIndicator;
     if Size.Y = 1 then case CtrlToArrow(Event.KeyCode) of
      kbLeft: ClickPart := sbLeftArrow;
      kbRight: ClickPart := sbRightArrow;
      kbCtrlLeft: ClickPart := sbPageLeft;
      kbCtrlRight: ClickPart := sbPageRight;
      kbHome: I := Min;
      kbEnd: I := Max;
      else Exit;
     end else case CtrlToArrow(Event.KeyCode) of
      kbUp: ClickPart := sbUpArrow;
      kbDown: ClickPart := sbDownArrow;
      kbPgUp: ClickPart := sbPageUp;
      kbPgDn: ClickPart := sbPageDown;
      kbCtrlPgUp: I := Min;
      kbCtrlPgDn: I := Max;
      else Exit;
     end;
     Clicked;
     if ClickPart <> sbIndicator then I := Value + ScrollStep(ClickPart);
     SetValue(I);
     ClearEvent(Event);
    end;
  end;
 end;

procedure TScrollBar.ScrollDraw;
 begin Message(Owner, evBroadcast, cmScrollBarChanged, @Self) end;

function TScrollBar.ScrollStep(Part: Integer): Integer;
 var Step: Integer;
 begin
  if Part and 2 = 0 then Step := ArStep else Step := PgStep;
  if Part and 1 = 0 then ScrollStep := -Step else ScrollStep := Step;
 end;

procedure TScrollBar.SetParams(AValue, AMin, AMax, APgStep,
  AArStep: Integer);
 var SValue: Integer;
 begin
  if AMax < AMin then AMax := AMin;
  if AValue < AMin then AValue := AMin;
  if AValue > AMax then AValue := AMax;
  SValue := Value;
  if (SValue <> AValue) or (Min <> AMin) or (Max <> AMax) then begin
   Value := AValue; Min := AMin; Max := AMax;
   DrawView;
   if SValue <> AValue then ScrollDraw;
  end;
  PgStep := APgStep;
  ArStep := AArStep;
 end;

procedure TScrollBar.SetRange(AMin, AMax: Integer);
 begin SetParams(Value, AMin, AMax, PgStep, ArStep) end;

procedure TScrollBar.SetStep(APgStep, AArStep: Integer);
 begin SetParams(Value, Min, Max, APgStep, AArStep) end;

procedure TScrollBar.SetValue(AValue: Integer);
 begin SetParams(AValue, Min, Max, PgStep, ArStep) end;

{ TScroller }

constructor TScroller.Init(var Bounds: TRect; AHScrollBar,
  AVScrollBar: PScrollBar);
 begin
  TView.Init(Bounds);
  Options := Options or ofSelectable;
  EventMask := EventMask or evBroadcast;
  HScrollBar := AHScrollBar;
  VScrollBar := AVScrollBar;
 end;

procedure TScroller.ChangeBounds(var Bounds: TRect);
 begin
  SetBounds(Bounds);
  Inc(DrawLock);
  SetLimit(Limit.X, Limit.Y);
  Dec(DrawLock);
  DrawFlag := False;
  DrawView;
 end;

procedure TScroller.CheckDraw;
 begin
  if (DrawLock = 0) and DrawFlag
   then begin DrawFlag := False; DrawView end;
 end;

function TScroller.GetPalette: PPalette;
 const P: String[Length(CScroller)] = CScroller;
 begin GetPalette := @P end;

procedure TScroller.HandleEvent(var Event: TEvent);
 begin
  TView.HandleEvent(Event);
  if (Event.What = evBroadcast) and (Event.Command = cmScrollBarChanged) and
     ((Event.InfoPtr = HScrollBar) or (Event.InfoPtr = VScrollBar))
   then ScrollDraw;
 end;

procedure TScroller.ScrollDraw;
 var D: TPoint;
 begin
  if HScrollBar <> nil then D.X := HScrollBar^.Value else D.X := 0;
  if VScrollBar <> nil then D.Y := VScrollBar^.Value else D.Y := 0;
  if (D.X <> Delta.X) or (D.Y <> Delta.Y) then begin
   SetCursor(Cursor.X + Delta.X - D.X, Cursor.Y + Delta.Y - D.Y);
   Delta := D;
   if DrawLock <> 0 then DrawFlag := True else DrawView;
  end;
 end;

procedure TScroller.ScrollTo(X, Y: Integer);
 begin
  Inc(DrawLock);
  if HScrollBar <> nil then HScrollBar^.SetValue(X);
  if VScrollBar <> nil then VScrollBar^.SetValue(Y);
  Dec(DrawLock);
  CheckDraw;
 end;

procedure TScroller.SetLimit(X, Y: Integer);
 begin
  Limit.X := X; Limit.Y := Y;
  Inc(DrawLock);
  if HScrollBar <> nil then
    HScrollBar^.SetParams(HScrollBar^.Value, 0, X - Size.X, Size.X - 1,
      HScrollBar^.ArStep);
  if VScrollBar <> nil then
    VScrollBar^.SetParams(VScrollBar^.Value, 0, Y - Size.Y, Size.Y - 1,
      VScrollBar^.ArStep);
  Dec(DrawLock);
  CheckDraw;
 end;

procedure TScroller.SetState(AState: Word; Enable: Boolean);
 procedure ShowSBar(SBar: PScrollBar);
  begin
   if (SBar <> nil)
    then if GetState(sfActive + sfSelected)
     then SBar^.Show else SBar^.Hide;
  end;
 begin
  TView.SetState(AState, Enable);
  if AState and (sfActive + sfSelected) <> 0
   then begin ShowSBar(HScrollBar); ShowSBar(VScrollBar) end;
 end;

{ TGroup }

constructor TGroup.Init(var Bounds: TRect);
 begin
  TView.Init(Bounds);
  Options := Options or (ofSelectable + ofBuffered);
  GetExtent(Clip);
  EventMask := $FFFF;
 end;

destructor TGroup.Done;
 var P, T: PView;
 begin
  Hide;
  P := Last;
  if P <> nil then begin
   repeat P^.Hide; P := P^.Prev until P = Last;
   repeat T := P^.Prev; Dispose(P, Done); P := T until Last = nil;
  end;
  FreeBuffer;
  TView.Done;
 end;

function TGroup.At(Index: Integer): PView;
 var temp:PView;
 begin
  temp := last;
  while Index>0 do begin dec(Index); temp:=temp^.Next end;
  At:=temp;
 end;

procedure TGroup.Awaken;
 procedure DoAwaken(P: PView); begin P^.Awaken end;
 begin ForEach(@DoAwaken) end;

procedure TGroup.ChangeBounds(var Bounds: TRect);
 var D:TPoint;
 procedure DoCalcChange(P: PView);
  var R: TRect;
  begin
   P^.CalcBounds(R, D);
   P^.ChangeBounds(R);
  end;
 begin
  D.X := Bounds.B.X - Bounds.A.X - Size.X;
  D.Y := Bounds.B.Y - Bounds.A.Y - Size.Y;
  if (D.X=0) and (D.Y=0) then begin
   SetBounds(Bounds); DrawView;
  end else begin
   FreeBuffer; SetBounds(Bounds); GetExtent(Clip); GetBuffer;
   Lock; ForEach(@DoCalcChange); Unlock;
  end;
 end;

function TGroup.DataSize: Word;
 var w:Word;
 procedure AddSubviewDataSize(P: PView);
  begin Inc(w, P^.DataSize) end;
 begin
  w:=0; ForEach(@AddSubviewDataSize); DataSize:=w;
 end;

procedure TGroup.Delete(P: PView);
 var SaveState: Word;
 begin
  SaveState := P^.State;
  P^.Hide;
  RemoveView(P);
  P^.Owner := nil;
  P^.Next := nil;
  if SaveState and sfVisible <> 0 then P^.Show;
 end;

procedure TGroup.Draw;
 var R: TRect;
 begin
  if Buffer = nil then begin
   GetBuffer;
   if Buffer <> nil then begin Inc(LockFlag); Redraw; Dec(LockFlag) end;
  end;
  if Buffer <> nil
   then WriteBuf(0, 0, Size.X, Size.Y, Buffer^)
   else begin GetClipRect(Clip); Redraw; GetExtent(Clip) end;
 end;

procedure TGroup.DrawSubViews(P, Bottom: PView);
 begin
  if P <> nil then while P <> Bottom do
   begin P^.DrawView; P := P^.NextView end;
 end;

procedure TGroup.EndModal(Command: Word);
 begin
  if State and sfModal <> 0
   then EndState := Command
   else TView.EndModal(Command);
 end;

procedure TGroup.EventError(var Event: TEvent);
 begin if Owner <> nil then Owner^.EventError(Event) end;

function TGroup.Execute: Word;
 var E: TEvent;
 begin
  repeat
   EndState := 0;
   repeat
    GetEvent(E);
    HandleEvent(E);
    if E.What <> evNothing then EventError(E);
   until EndState <> 0;
  until Valid(EndState);
  Execute := EndState;
 end;

function TGroup.ExecView(P: PView): Word;
 var
  SaveOptions: Word;
  SaveOwner: PGroup;
  SaveTopView: PView;
  SaveCurrent: PView;
  SaveCommands: TCommandSet;
 begin
  if P <> nil then begin
   SaveOptions := P^.Options;
   SaveOwner := P^.Owner;
   SaveTopView := TheTopView;
   SaveCurrent := Current;
   GetCommands(SaveCommands);
   TheTopView := P;
   P^.Options := P^.Options and not ofSelectable;
   P^.SetState(sfModal, True);
   SetCurrent(P, EnterSelect);
   if SaveOwner = nil then Insert(P);
   ExecView := P^.Execute;
   if SaveOwner = nil then Delete(P);
   SetCurrent(SaveCurrent, LeaveSelect);
   P^.SetState(sfModal, False);
   P^.Options := SaveOptions;
   TheTopView := SaveTopView;
   SetCommands(SaveCommands);
  end else ExecView := cmCancel;
 end;

function TGroup.First: PView;
 begin if Last = nil then First := nil else First := Last^.Next end;

function TGroup.FirstMatch(AState: Word; AOptions: Word): PView;
 function Matches(P: PView): Boolean;
  begin
   Matches:=
    (P^.State and AState = AState) and
    (P^.Options and AOptions = AOptions);
  end;
 begin FirstMatch:=FirstThat(@Matches) end;

FUNCTION TGroup.FirstThat (P: Pointer): PView;
{$IFDEF BIT_32}
 TYPE NewFuncPtr = FUNCTION (_BP:longint; AView: PView): Boolean;
 VAR Hold_BP: longint; Tp: PView;
{$ELSE}
 TYPE NewFuncPtr = FUNCTION (AView: PView; _BP: Word): Boolean;
 VAR Hold_BP: Word; Tp: PView;
{$ENDIF}
 BEGIN
   {$IFDEF BIT_16}                                    { 16 BIT CODE }
   ASM
     MOV AX, [BP];                                    { Load AX from BP }
     MOV [Hold_BP], AX;                               { Hold value }
   END;
   {$ELSE}
   ASM
     MOVL (%EBP), %EAX;
     MOVL %EAX, Hold_BP;                               { Hold value }
   END;
   {$ENDIF}
   If (Last<>Nil) Then Begin
    Tp := Last;                                      { Set temporary ptr }
    Repeat
     Tp := Tp^.Next;                                { Get next view }
     {$IFDEF BIT_32}
     If NewFuncPtr(P)(Hold_Bp, Tp) Then Begin  { Test each view }
     {$ELSE}
     If NewFuncPtr(P)(Tp, Hold_Bp) Then Begin       { Test each view }
     {$ENDIF}
      FirstThat := Tp;                             { View returned true }
      Exit;                                        { Now exit }
     End;
    Until (Tp=Last);                                 { Until last }
    FirstThat := Nil;                                { None passed test }
   End Else FirstThat := Nil;                         { Return nil }
 END;

function TGroup.FindNext(Forwards: Boolean): PView;
 var P: PView;
 begin
  FindNext := nil;
  if Current <> nil then begin
   P := Current;
   repeat
    if Forwards then P := P^.Next else P := P^.Prev;
   until
    ((P^.State and (sfVisible + sfDisabled) = sfVisible) and
    (P^.Options and ofSelectable <> 0)) or (P = Current);
   if P <> Current then FindNext:=P;
  end;
 end;

function TGroup.FocusNext(Forwards: Boolean): Boolean;
 var P: PView;
 begin
  P := FindNext(Forwards);
  FocusNext := True;
  if P <> nil then FocusNext := P^.Focus;
 end;

PROCEDURE TGroup.ForEach (P: Pointer);
{$IFDEF BIT_32}
 TYPE NewProcPtr = PROCEDURE (_BP: longint; AView: PView);
 VAR Hold_BP: longint; Tp,Hp,L0: PView;
{$ELSE}
 TYPE NewProcPtr = PROCEDURE (AView: PView; _BP: Word);
 VAR Hold_BP: Word; Tp,Hp,L0: PView;
{$ENDIF}
{ Vars Hp and L0 are necessary to hold original pointers in case   }
{ when some view closes himself as a result of broadcast message ! }
 BEGIN
   {$IFDEF BIT_16}                                    { 16 BIT CODE }
   ASM
     MOV AX, [BP];                                    { Load AX from BP }
     MOV [Hold_BP], AX;                               { Hold value }
   END;
   {$ELSE}
   ASM
     MOVL (%EBP), %EAX;
     MOVL %EAX, Hold_BP;                               { Hold value }
   END;
   {$ENDIF}
   If (Last<>Nil) Then Begin
     Tp := Last; Hp:=Tp^.Next; L0:=Last;              { Set temporary ptr }
     Repeat
       Tp := Hp; Hp:=Tp^.Next;                        { Get next view }
       {$IFDEF BIT_32}
       NewProcPtr(P)(Hold_BP, Tp);                    { Do each view }
       {$ELSE}
       NewProcPtr(P)(Tp, Hold_BP);                    { Do each view }
       {$ENDIF}
     Until (Tp=L0);                                   { Until last }
   End;
 END;

procedure TGroup.FreeBuffer;
 begin
  if (Options and ofBuffered <> 0) and (Buffer <> nil)
   then begin FreeMem(Buffer,(Size.X*Size.Y) shl 1); Buffer:=Nil end;
 end;

procedure TGroup.GetBuffer;
 begin
  if (state and sfExposed)<>0
   then if ((options and ofBuffered)<>0) and (buffer=Nil)
    then GetMem(buffer,(size.x * size.y) shl 1);
 end;

procedure TGroup.GetData(var Rec);
 type Bytes = array[0..65534] of Byte;
 var I: Word; V: PView;
 begin
  I := 0;
  if Last <> nil then begin
   V := Last;
   repeat
    V^.GetData(Bytes(Rec)[I]);
    Inc(I, V^.DataSize);
    V := V^.Prev;
   until V = Last;
  end;
 end;

function TGroup.GetHelpCtx: Word;
 var H: Word;
 begin
  H:= hcNoContext;
  if Current <> nil then H := Current^.GetHelpCtx;
  if H = hcNoContext then H := TView.GetHelpCtx;
  GetHelpCtx := H;
 end;

procedure TGroup.HandleEvent(var Event: TEvent);
 procedure DoHandleEvent(P: PView);
  begin
   if (P = nil) or ((P^.State and sfDisabled <> 0)
    and (Event.What and (PositionalEvents or FocusedEvents) <> 0)) then Exit;
   case Phase of
    phPreProcess: if P^.Options and ofPreProcess = 0 then Exit;
    phPostProcess: if P^.Options and ofPostProcess = 0 then Exit;
   end;
   if Event.What and P^.EventMask <> 0 then P^.HandleEvent(Event);
  end;
 function ContainsMouse(P: PView): Boolean;
  begin
   ContainsMouse:=
    (P^.State and sfVisible <> 0) and P^.MouseInView(Event.Where);
  end;
 begin
  TView.HandleEvent(Event);
  if Event.What and FocusedEvents <> 0 then begin
   Phase := phPreProcess;
   ForEach(@DoHandleEvent);
   Phase := phFocused;
   DoHandleEvent(Current);
   Phase := phPostProcess;
   ForEach(@DoHandleEvent);
  end else begin
   Phase := phFocused;
   if (Event.What and PositionalEvents <> 0)
    then DoHandleEvent(FirstThat(@ContainsMouse))
    else ForEach(@DoHandleEvent);
  end;
 end;

function TGroup.IndexOf(P: PView): Integer;
 var temp:PView; index:integer;
 begin
  index:=0;
  if Last<>Nil then begin
   temp:=Last;
   repeat inc(index); temp:=temp^.Next until (temp=P) or (temp=Last);
   if temp<>p then index:=0;
  end;
  IndexOf:=index;
 end;

procedure TGroup.Insert(P: PView);
 begin InsertBefore(P, First) end;

procedure TGroup.InsertBefore(P, Target: PView);
 var SaveState: Word;
 begin
  if (P <> nil) and (P^.Owner = nil) and
    ((Target = nil) or (Target^.Owner = @Self)) then begin
   if P^.Options and ofCenterX <> 0
    then P^.Origin.X := (Size.X - P^.Size.X) div 2;
   if P^.Options and ofCenterY <> 0
    then P^.Origin.Y := (Size.Y - P^.Size.Y) div 2;
   SaveState := P^.State;
   P^.Hide;
   InsertView(P, Target);
   if SaveState and sfVisible <> 0 then P^.Show;
   if State and sfActive <> 0 then P^.SetState(sfActive, True);
  end;
 end;

procedure TGroup.InsertView(P, Target: PView);
 begin
  P^.Owner := @Self;
  if Target <> nil then begin
    Target := Target^.Prev;
    P^.Next := Target^.Next;
    Target^.Next := P;
  end else begin
   if Last = nil
    then P^.Next := P
    else begin P^.Next := Last^.Next; Last^.Next := P end;
   Last := P;
  end;
 end;

procedure TGroup.Lock;
 begin
  if (Buffer <> nil) or (LockFlag <> 0) then Inc(LockFlag);
 end;

procedure TGroup.Redraw;
 begin DrawSubViews(First,nil) end;

procedure TGroup.RemoveView(P: PView);
 var cur:PView;
 begin if Last<>Nil then begin
  cur:=Last;
  while true do begin
   if p=cur^.Next then begin
    cur^.Next:=P^.Next;
    if Last=P then
     begin if cur^.Next=P then Last:=Nil else Last:=cur; break end;
   end;
   if cur^.next=Last then break;
   cur:=cur^.Next;
  end;
 end end;

procedure TGroup.ResetCurrent;
 begin
  SetCurrent(FirstMatch(sfVisible, ofSelectable), NormalSelect);
 end;

procedure TGroup.ResetCursor;
 begin if Current <> nil then Current^.ResetCursor end;

procedure TGroup.SelectNext(Forwards: Boolean);
 var P: PView;
 begin
  P := FindNext(Forwards);
  if P <> nil then P^.Select;
 end;

procedure TGroup.SetCurrent(P: PView; Mode: SelectMode);
 procedure SelectView(P: PView; Enable: Boolean);
  begin
   if P <> nil then P^.SetState(sfSelected, Enable);
  end;
 procedure FocusView(P: PView; Enable: Boolean);
  begin
   if (State and sfFocused <> 0) and (P <> nil)
    then P^.SetState(sfFocused, Enable);
  end;
 begin
  if Current <> P then begin
   Lock;
   FocusView(Current, False);
   if Mode <> EnterSelect then SelectView(Current, False);
   if Mode <> LeaveSelect then SelectView(P, True);
   FocusView(P, True);
   Current := P;
   Unlock;
  end;
 end;

procedure TGroup.SetData(var Rec);
 type Bytes = array[0..65534] of Byte;
 var I: Word; V: PView;
 begin
  I := 0;
  if Last <> nil then begin
   V := Last;
   repeat
    V^.SetData(Bytes(Rec)[I]);
    Inc(I, V^.DataSize);
    V := V^.Prev;
   until V = Last;
  end;
 end;

procedure TGroup.SetState(AState: Word; Enable: Boolean);
 procedure DoSetState(P: PView);
  begin P^.SetState(AState,Enable) end;
 procedure DoExpose(P: PView);
  begin
   if P^.State and sfVisible <> 0 then P^.SetState(sfExposed, Enable);
  end;
 begin
  TView.SetState(AState, Enable);
  case AState of
   sfActive, sfDragging:begin Lock; ForEach(@DoSetState); Unlock end;
   sfFocused:if Current <> nil then Current^.SetState(sfFocused, Enable);
   sfExposed:begin ForEach(@DoExpose); if not Enable then FreeBuffer end;
  end;
 end;

procedure TGroup.Unlock;
 begin
  if LockFlag <> 0
   then begin Dec(LockFlag); if LockFlag = 0 then DrawView end;
 end;

function TGroup.Valid(Command: Word): Boolean;
 function IsInvalid(P: PView): Boolean;
  begin IsInvalid := not P^.Valid(Command) end;
 begin
  Valid := True;
  if Command = cmReleasedFocus
   then begin
    if (Current <> nil) and (Current^.Options and ofValidate <> 0)
     then Valid := Current^.Valid(Command);
   end
   else Valid := FirstThat(@IsInvalid) = nil;
 end;

{ TWindow }

constructor TWindow.Init(var Bounds: TRect; ATitle: TTitleStr;
  ANumber: Integer);
 begin
  TGroup.Init(Bounds);
  State := State or sfShadow;
  Options := Options or (ofSelectable + ofTopSelect);
  GrowMode := gfGrowAll + gfGrowRel;
  Flags := wfMove + wfGrow + wfClose + wfZoom;
  Title := NewStr(ATitle);
  Number := ANumber;
  Palette := wpBlueWindow;
  InitFrame;
  if Frame <> nil then Insert(Frame);
  GetBounds(ZoomRect);
 end;

destructor TWindow.Done;
 begin TGroup.Done; DisposeStr(Title) end;

procedure TWindow.Close;
 begin if Valid(cmClose) then Free end;

function TWindow.GetPalette: PPalette;
 const
  P: array[wpBlueWindow..wpGrayWindow] of string[Length(CBlueWindow)] =
    (CBlueWindow, CCyanWindow, CGrayWindow);
 begin GetPalette:=@P[Palette] end;

function TWindow.GetTitle(MaxSize: Integer): TTitleStr;
 begin
  if Title <> nil then GetTitle := Title^ else GetTitle := '';
 end;

procedure TWindow.HandleEvent(var Event: TEvent);
 var Limits: TRect; Min, Max: TPoint;
 begin
  TGroup.HandleEvent(Event);
  if (Event.What = evCommand) then case Event.Command of
   cmResize:if Flags and (wfMove + wfGrow) <> 0 then begin
     Owner^.GetExtent(Limits);
     SizeLimits(Min, Max);
     DragView(Event, DragMode or (Flags and (wfMove + wfGrow)),
      Limits, Min, Max);
     ClearEvent(Event);
    end;
   cmClose:
    if (Flags and wfClose <> 0) and
       ((Event.InfoPtr = nil) or (Event.InfoPtr = @Self)) then begin
     ClearEvent(Event);
     if State and sfModal = 0
      then Close
      else begin
       Event.What := evCommand;
       Event.Command := cmCancel;
       PutEvent(Event);
       ClearEvent(Event);
      end;
    end;
   cmZoom:
    if (Flags and wfZoom <> 0) and
       ((Event.InfoPtr = nil) or (Event.InfoPtr = @Self))
     then begin Zoom; ClearEvent(Event) end;
  end else if Event.What = evKeyDown then case Event.KeyCode of
   kbTab: begin FocusNext(False); ClearEvent(Event) end;
   kbShiftTab: begin FocusNext(True); ClearEvent(Event) end;
  end else if (Event.What = evBroadcast) and
    (Event.Command = cmSelectWindowNum) and
    (Event.InfoInt = Number) and
    (Options and ofSelectable <> 0)
   then begin Select; ClearEvent(Event) end;
 end;

procedure TWindow.InitFrame;
 var R: TRect;
 begin GetExtent(R); Frame := New(PFrame, Init(R)) end;

procedure TWindow.SetState(AState: Word; Enable: Boolean);
 var WindowCommands: TCommandSet;
 begin
  TGroup.SetState(AState, Enable);
  if AState = sfSelected then SetState(sfActive, Enable);
  if (AState = sfSelected) or ((AState = sfExposed) and
    (State and sfSelected <> 0))
   then begin
{$IFDEF OS_Linux}
    WindowCommands := TCommandSet([cmNext, cmPrev]);
    if Flags and wfGrow + wfMove <> 0
     then WindowCommands := WindowCommands + TCommandSet([cmResize]);
    if Flags and wfClose <> 0
     then WindowCommands := WindowCommands + TCommandSet([cmClose]);
    if Flags and wfZoom <> 0
     then WindowCommands := WindowCommands + TCommandSet([cmZoom]);
{$ELSE}
    WindowCommands := [cmNext, cmPrev];
    if Flags and wfGrow + wfMove <> 0
     then WindowCommands := WindowCommands + [cmResize];
    if Flags and wfClose <> 0
     then WindowCommands := WindowCommands + [cmClose];
    if Flags and wfZoom <> 0
     then WindowCommands := WindowCommands + [cmZoom];
{$ENDIF}
    if Enable
     then EnableCommands(WindowCommands)
     else DisableCommands(WindowCommands);
   end;
 end;

function TWindow.StandardScrollBar(AOptions: Word): PScrollBar;
 var R: TRect; S: PScrollBar;
 begin
  GetExtent(R);
  if AOptions and sbVertical = 0
   then R.Assign(R.A.X + 2, R.B.Y-1, R.B.X-2, R.B.Y)
   else R.Assign(R.B.X-1,R.A.Y+1,R.B.X,R.B.Y-1);
  S := New(PScrollBar, Init(R));
  Insert(S);
  if AOptions and sbHandleKeyboard <> 0
   then S^.Options := S^.Options or ofPostProcess;
  StandardScrollBar := S;
 end;

procedure TWindow.SizeLimits(var Min, Max: TPoint);
 begin
  TView.SizeLimits(Min, Max);
  Min := MinWinSize;
 end;

procedure TWindow.Zoom;
 var R: TRect; Max, Min: TPoint;
 begin
  SizeLimits(Min, Max);
  if (Size.X<>Max.X) or (Size.y<>Max.Y) then begin
   GetBounds(ZoomRect);
   R.A.X := 0; R.A.Y:=0;
   R.B := Max;
   Locate(R);
  end else Locate(ZoomRect);
 end;

function Message(Receiver: PView; What, Command: Word;
  InfoPtr: Pointer): Pointer;
 var Event: TEvent;
 begin
  Message := nil;
  if Receiver <> nil then begin
   Event.What := What;
   Event.Command := Command;
   Event.InfoPtr := InfoPtr;
   Receiver^.HandleEvent(Event);
   if Event.What = evNothing then Message := Event.InfoPtr;
  end;
 end;

begin
{$IFDEF OS_Linux}
  with MinWinSize do begin X:=16; Y:=6 end;
  with ShadowSize do begin X:=2; Y:=1 end;
{$ENDIF}
end.

